import logging

from .pi30 import pi30

log = logging.getLogger("pi30max")

QUERY_COMMANDS = {
    "QSID": {
        "name": "QSID",
        "description": "Device Serial Number inquiry",
        "help": " -- queries the device serial number (length greater than 14)",
        "type": "QUERY",
        "response_type": "INDEXED",
        "response": [[1, "Serial Number", "bytes.decode:r[2:int(r[0:2])+2]", "", {"icon": "mdi:identifier"}]],
        "test_responses": [
            b"(1492932105105335005535\x94\x0e\r",
        ],
    },
    "QVFW3": {
        "name": "QVFW3",
        "description": "Remote CPU firmware version inquiry",
        "type": "QUERY",
        "response_type": "INDEXED",
        "response": [[1, "Remote CPU firmware version", "bytes.decode", "", {"icon": "mdi:identifier"}]],
        "test_responses": [],
    },
    "VERFW": {
        "name": "VERFW",
        "description": "Bluetooth version inquiry",
        "type": "QUERY",
        "response_type": "INDEXED",
        "response": [[1, "Bluetooth version", "bytes.decode", "", {"icon": "mdi:identifier"}]],
        "test_responses": [],
    },
    "QPIRI": {
        "name": "QPIRI",
        "description": "Current Settings inquiry",
        "type": "QUERY",
        "response_type": "INDEXED",
        "response": [
            [
                1,
                "AC Input Voltage",
                "float",
                "V",
                {"icon": "mdi:transmission-tower-import", "device-class": "voltage"},
            ],
            [2, "AC Input Current", "float", "A", {"icon": "mdi:current-ac", "device-class": "current"}],
            [
                3,
                "AC Output Voltage",
                "float",
                "V",
                {"icon": "mdi:transmission-tower-export", "device-class": "voltage"},
            ],
            [4, "AC Output Frequency", "float", "Hz", {"icon": "mdi:current-ac", "device-class": "frequency"}],
            [5, "AC Output Current", "float", "A", {"icon": "mdi:current-ac", "device-class": "current"}],
            [6, "AC Output Apparent Power", "int", "VA", {"icon": "mdi:power-plug", "device-class": "apparent_power"}],
            [
                7,
                "AC Output Active Power",
                "int",
                "W",
                {"icon": "mdi:power-plug", "device-class": "power", "state_class": "measurement"},
            ],
            [8, "Battery Voltage", "float", "V", {"icon": "mdi:battery-outline", "device-class": "voltage"}],
            [9, "Battery Recharge Voltage", "float", "V", {"icon": "mdi:battery-outline", "device-class": "voltage"}],
            [10, "Battery Under Voltage", "float", "V", {"icon": "mdi:battery-outline", "device-class": "voltage"}],
            [
                11,
                "Battery Bulk Charge Voltage",
                "float",
                "V",
                {"icon": "mdi:battery-outline", "device-class": "voltage"},
            ],
            [
                12,
                "Battery Float Charge Voltage",
                "float",
                "V",
                {"icon": "mdi:battery-outline", "device-class": "voltage"},
            ],
            [
                13,
                "Battery Type",
                "option",
                [
                    "AGM",
                    "Flooded",
                    "User",
                    "Pylontech",
                    "Shinheung",
                    "WECO",
                    "Soltaro",
                    "LIb-protocol compatible",
                    "3rd party Lithium",
                ],
            ],
            [14, "Max AC Charging Current", "int", "A", {"icon": "mdi:current-ac", "device-class": "current"}],
            [15, "Max Charging Current", "int", "A", {"icon": "mdi:current-ac", "device-class": "current"}],
            [16, "Input Voltage Range", "option", ["Appliance", "UPS"]],
            [
                17,
                "Output Source Priority",
                "option",
                [
                    "Utility Solar Battery",
                    "Solar Utility Battery",
                    "Solar Battery Utility",
                ],
            ],
            [
                18,
                "Charger Source Priority",
                "option",
                [
                    "Utility first",
                    "Solar first",
                    "Solar + Utility",
                    "Only solar charging permitted",
                ],
            ],
            [19, "Max Parallel Units", "int", "units"],
            [
                20,
                "Machine Type",
                "str_keyed",
                {"00": "Grid tie", "01": "Off Grid", "10": "Hybrid"},
            ],
            [21, "Topology", "option", ["transformerless", "transformer"]],
            [
                22,
                "Output Mode",
                "option",
                [
                    "single machine output",
                    "parallel output",
                    "Phase 1 of 3 Phase output",
                    "Phase 2 of 3 Phase output",
                    "Phase 3 of 3 Phase output",
                    "Phase 1 of 2 phase output",
                    "Phase 2 of 2 phase output (120°)",
                    "Phase 2 of 2 phase output (180°)",
                    "unknown output",
                ],
            ],
            [23, "Battery Redischarge Voltage", "float", "V"],
            [
                24,
                "PV OK Condition",
                "option",
                [
                    "As long as one unit of inverters has connect PV, parallel system will consider PV OK",
                    "Only All of inverters have connect PV, parallel system will consider PV OK",
                ],
            ],
            [
                25,
                "PV Power Balance",
                "option",
                [
                    "PV input max current will be the max charged current",
                    "PV input max power will be the sum of the max charged power and loads power",
                ],
            ],
            [26, "Max charging time for CV stage", "int", "min"],
            [
                27,
                "Operation Logic",
                "option",
                ["Automatic mode", "On-line mode", "ECO mode"],
            ],
            [28, "Max discharging current", "int", "A", {"icon": "mdi:current-ac", "device-class": "current"}],
        ],
        "test_responses": [
            b"(230.0 21.7 230.0 50.0 21.7 5000 4000 48.0 46.0 42.0 56.4 54.0 0 10 010 1 0 0 6 01 0 0 54.0 0 1\x6F\x7E\r",
        ],
    },
    "QFLAG": {
        "name": "QFLAG",
        "description": "Flag Status inquiry",
        "help": " -- queries the enabled / disabled state of various Inverter settings (e.g. buzzer, overload, interrupt alarm)",
        "type": "QUERY",
        "response": [
            [
                "enflags",
                "Device Status",
                {
                    "a": {"name": "Buzzer", "state": "disabled"},
                    "b": {"name": "Overload Bypass", "state": "disabled"},
                    "d": {"name": "Solar Feed to Grid", "state": "disabled"},
                    "k": {"name": "LCD Reset to Default", "state": "disabled"},
                    "u": {"name": "Overload Restart", "state": "disabled"},
                    "v": {"name": "Over Temperature Restart", "state": "disabled"},
                    "x": {"name": "LCD Backlight", "state": "disabled"},
                    "y": {
                        "name": "Primary Source Interrupt Alarm",
                        "state": "disabled",
                    },
                    "z": {"name": "Record Fault Code", "state": "disabled"},
                },
            ]
        ],
        "test_responses": [],
    },
    "QPIGS": {
        "name": "QPIGS",
        "description": "General Status Parameters inquiry",
        "type": "QUERY",
        "response_type": "INDEXED",
        "response": [
            [
                1,
                "AC Input Voltage",
                "float",
                "V",
                {"icon": "mdi:transmission-tower-export", "device-class": "voltage"},
            ],
            [2, "AC Input Frequency", "float", "Hz", {"icon": "mdi:current-ac", "device-class": "frequency"}],
            [3, "AC Output Voltage", "float", "V", {"icon": "mdi:power-plug", "device-class": "voltage"}],
            [4, "AC Output Frequency", "float", "Hz", {"icon": "mdi:current-ac", "device-class": "frequency"}],
            [5, "AC Output Apparent Power", "int", "VA", {"icon": "mdi:power-plug", "device-class": "apparent_power"}],
            [
                6,
                "AC Output Active Power",
                "int",
                "W",
                {"icon": "mdi:power-plug", "device-class": "power", "state_class": "measurement"},
            ],
            [7, "AC Output Load", "int", "%", {"icon": "mdi:brightness-percent"}],
            [8, "BUS Voltage", "int", "V", {"icon": "mdi:details", "device-class": "voltage"}],
            [9, "Battery Voltage", "float", "V", {"icon": "mdi:battery-outline", "device-class": "voltage"}],
            [10, "Battery Charging Current", "int", "A", {"icon": "mdi:current-dc", "device-class": "current"}],
            [11, "Battery Capacity", "int", "%", {"device-class": "battery"}],
            [
                12,
                "Inverter Heat Sink Temperature",
                "int",
                "°C",
                {"icon": "mdi:details", "device-class": "temperature"},
            ],
            [13, "PV1 Input Current", "float", "A", {"icon": "mdi:solar-power", "device-class": "current"}],
            [14, "PV1 Input Voltage", "float", "V", {"icon": "mdi:solar-power", "device-class": "voltage"}],
            [
                15,
                "Battery Voltage from SCC",
                "float",
                "V",
                {"icon": "mdi:battery-outline", "device-class": "voltage"},
            ],
            [
                16,
                "Battery Discharge Current",
                "int",
                "A",
                {"icon": "mdi:battery-negative", "device-class": "current"},
            ],
            [
                17,
                "Device Status",
                "flags",
                [
                    "Is SBU Priority Version Added",
                    "Is Configuration Changed",
                    "Is SCC Firmware Updated",
                    "Is Load On",
                    "Is Battery Voltage to Steady While Charging",
                    "Is Charging On",
                    "Is SCC Charging On",
                    "Is AC Charging On",
                ],
            ],
            [18, "Battery Voltage Offset for Fans On", "int", "10mV"],
            [19, "EEPROM Version", "int", ""],
            [
                20,
                "PV1 Charging Power",
                "int",
                "W",
                {"icon": "mdi:solar-power", "device-class": "power", "state_class": "measurement"},
            ],
            [
                21,
                "Device Status2",
                "flags",
                ["Is Charging to Float", "Is Switched On", "Is Dustproof Installed"],
            ],
            [22, "Solar Feed to Grid", "option", ["Disabled", "Enabled"]],
            [
                23,
                "Country",
                "keyed",
                {
                    "00": "India",
                    "01": "Germany",
                    "02": "South America",
                },
            ],
            [
                24,
                "Solar Feed to Grid Power",
                "int",
                "W",
                {"icon": "mdi:solar-power", "device-class": "power", "state_class": "measurement"},
            ],
        ],
        "test_responses": [
            b"(227.2 50.0 230.3 50.0 0829 0751 010 447 54.50 020 083 0054 02.7 323.6 00.00 00000 00010110 00 00 00879 010\xf1\x8c\r",
        ],
    },
    "QPIGS2": {
        "name": "QPIGS2",
        "description": "General Status Parameters inquiry 2",
        "type": "QUERY",
        "response_type": "INDEXED",
        "response": [
            [1, "PV2 Input Current", "float", "A", {"icon": "mdi:solar-power", "device-class": "current"}],
            [2, "PV2 Input Voltage", "float", "V", {"icon": "mdi:solar-power", "device-class": "voltage"}],
            [
                3,
                "PV2 Charging Power",
                "int",
                "W",
                {"icon": "mdi:solar-power", "device-class": "power", "state_class": "measurement"},
            ],
        ],
        "test_responses": [
            b"(03.1 327.3 01026 \xc9\x8b\r",
        ],
    },
    "QPGS": {
        "name": "QPGS",
        "description": "Parallel Information inquiry",
        "help": " -- example: QPGS0 queries the values of various metrics from instance 0 of parallel setup Inverters",
        "type": "QUERY",
        "response_type": "INDEXED",
        "response": [
            [1, "Parallel instance number", "option", ["Not valid", "valid"]],
            [2, "Serial number", "bytes:r.decode()", "", {"icon": "mdi:identifier"}],
            [
                3,
                "Work mode",
                "str_keyed",
                {
                    "P": "Power On Mode",
                    "S": "Standby Mode",
                    "L": "Line Mode",
                    "B": "Battery Mode",
                    "F": "Fault Mode",
                    "H": "Power Saving Mode",
                    "D": "Shutdown Mode",
                    "Y": "Bypass",
                },
            ],
            [
                4,
                "Fault code",
                "str_keyed",
                {
                    "00": "No fault",
                    "01": "Fan is locked",
                    "02": "Over temperature",
                    "03": "Battery voltage is too high",
                    "04": "Battery voltage is too low",
                    "05": "Output short circuited or Over temperature",
                    "06": "Output voltage is too high",
                    "07": "Over load time out",
                    "08": "Bus voltage is too high",
                    "09": "Bus soft start failed",
                    "10": "PV over current",
                    "11": "PV over voltage",
                    "12": "DC over current",
                    "13": "Battery discharge over current",
                    "51": "Over current inverter",
                    "52": "Bus voltage too low",
                    "53": "Inverter soft start failed",
                    "54": "Self-test failed",
                    "55": "Over DC voltage on output of inverter",
                    "56": "Battery connection is open",
                    "57": "Current sensor failed",
                    "58": "Output voltage is too low",
                    "60": "Power feedback protection",
                    "71": "Firmware version different",
                    "72": "Current sharing fault",
                    "80": "CAN communication failed",
                    "81": "Parallel host line lost",
                    "82": "Parallel synchronized signal lost",
                    "83": "Parallel battery voltage detect different",
                    "84": "AC input voltage or frequency detected different",
                    "85": "AC output current unbalanced",
                    "86": "AC output mode setting different",
                },
            ],
            [5, "Grid Voltage", "float", "V", {"icon": "mdi:power-plug", "device-class": "voltage"}],
            [6, "Grid Frequency", "float", "Hz", {"icon": "mdi:current-ac", "device-class": "frequency"}],
            [7, "AC Output Voltage", "float", "V", {"icon": "mdi:power-plug", "device-class": "voltage"}],
            [8, "AC Output Frequency", "float", "Hz", {"icon": "mdi:current-ac", "device-class": "frequency"}],
            [9, "AC Output Apparent Power", "int", "VA", {"icon": "mdi:power-plug", "device-class": "apparent_power"}],
            [
                10,
                "AC Output Active Power",
                "int",
                "W",
                {"icon": "mdi:power-plug", "device-class": "power", "state_class": "measurement"},
            ],
            [11, "Load Percentage", "int", "%", {"icon": "mdi:brightness-percent"}],
            [12, "Battery Voltage", "float", "V", {"icon": "mdi:battery-outline", "device-class": "voltage"}],
            [13, "Battery Charging Current", "int", "A", {"icon": "mdi:current-dc", "device-class": "current"}],
            [14, "Battery Capacity", "int", "%", {"device-class": "battery"}],
            [15, "PV1 Input Voltage", "float", "V", {"icon": "mdi:solar-power", "device-class": "voltage"}],
            [
                16,
                "Total Charging Current",
                "int",
                "A",
                {"icon": "mdi:brightness-percent", "device-class": "current"},
            ],
            [
                17,
                "Total AC Output Apparent Power",
                "int",
                "VA",
                {"icon": "mdi:power-plug", "device-class": "apparent_power"},
            ],
            [
                18,
                "Total Output Active Power",
                "int",
                "W",
                {"icon": "mdi:power-plug", "device-class": "power", "state_class": "measurement"},
            ],
            [
                19,
                "Total AC Output Percentage",
                "int",
                "%",
                {"icon": "mdi:brightness-percent"},
            ],
            [
                20,
                "Inverter Status",
                "flags",
                [
                    "Is SCC OK",
                    "Is AC Charging",
                    "Is SCC Charging",
                    "Is Battery Over Voltage",
                    "Is Battery Under Voltage",
                    "Is Line Lost",
                    "Is Load On",
                    "Is Configuration Changed",
                ],
            ],
            [
                21,
                "Output mode",
                "option",
                [
                    "single machine",
                    "parallel output",
                    "Phase 1 of 3 phase output",
                    "Phase 2 of 3 phase output",
                    "Phase 3 of 3 phase output",
                    "Phase 1 of 2 phase output",
                    "Phase 2 of 2 phase output (120°)",
                    "Phase 2 of 2 phase output (180°)",
                    "Unknown Output Mode",
                ],
            ],
            [
                22,
                "Charger source priority",
                "option",
                ["Utility first", "Solar first", "Solar + Utility", "Solar only"],
            ],
            [23, "Max Charger Current", "int", "A", {"device-class": "current"}],
            [24, "Max Charger Range", "int", "A", {"device-class": "current"}],
            [25, "Max AC Charger Current", "int", "A", {"device-class": "current"}],
            [26, "PV1 Input Current", "int", "A", {"icon": "mdi:solar-power", "device-class": "power"}],
            [
                27,
                "Battery Discharge Current",
                "int",
                "A",
                {"icon": "mdi:battery-negative", "device-class": "current"},
            ],
            [28, "PV2 Input Voltage", "float", "V", {"icon": "mdi:solar-power", "device-class": "voltage"}],
            [29, "PV2 Input Current", "int", "A", {"icon": "mdi:solar-power", "device-class": "current"}],
        ],
        "test_responses": [
            b"(0 92932105105315 B 00 000.0 00.00 230.0 50.00 0989 0907 012 53.2 009 090 349.8 009 00989 00907 011 10100110 0 1 100 120 030 02 000 275.3 02i]\r",
        ],
        "regex": "QPGS(\\d+)$",
    },
    "QMOD": {
        "name": "QMOD",
        "description": "Mode inquiry",
        "type": "QUERY",
        "response_type": "INDEXED",
        "response": [
            [
                1,
                "Device Mode",
                "str_keyed",
                {
                    "P": "Power on",
                    "S": "Standby",
                    "L": "Line",
                    "B": "Battery",
                    "F": "Fault",
                    "H": "Power Saving",
                    "D": "Shutdown",
                },
            ]
        ],
        "test_responses": [
            b"(S\xe5\xd9\r",
            b"(B\xe7\xc9\r",
        ],
    },
    "QPIWS": {
        "name": "QPIWS",
        "description": "Warning status inquiry",
        "type": "QUERY",
        "response": [
            [
                "stat_flags",
                "Warning",
                [
                    "PV loss warning",
                    "Inverter fault",
                    "Bus over fault",
                    "Bus under fault",
                    "Bus soft fail fault",
                    "Line fail warning",
                    "OPV short warning",
                    "Inverter voltage too low fault",
                    "Inverter voltage too high fault",
                    "Over temperature fault",
                    "Fan locked fault",
                    "Battery voltage too high fault",
                    "Battery low alarm warning",
                    "Reserved",
                    "Battery under shutdown warning",
                    "Battery derating warning",
                    "Overload fault",
                    "EEPROM fault",
                    "Inverter over current fault",
                    "Inverter soft fail fault",
                    "Self test fail fault",
                    "OP DC voltage over fault",
                    "Battery open fault",
                    "Current sensor fail fault",
                    "Battery short fault",
                    "Power limit warning",
                    "PV voltage high warning",
                    "MPPT overload fault",
                    "MPPT overload warning",
                    "Battery too low to charge warning",
                    "",
                    "Battery weak",
                    "Battery weak",
                    "Battery weak",
                    "",
                    "Battery equalisation warning",
                ],
            ]
        ],
        "test_responses": [
            b"(00000100000000001000000000000000\x56\xA6\r",
            b"(000000000000000000000000000000000000<\x8e\r",
        ],
    },
    "QDI": {
        "name": "QDI",
        "description": "Default Settings inquiry",
        "type": "QUERY",
        "response": [
            ["float", "AC Output Voltage", "V"],
            ["float", "AC Output Frequency", "Hz"],
            ["int", "Max AC Charging Current", "A"],
            ["float", "Battery Under Voltage", "V"],
            ["float", "Battery Float Charge Voltage", "V"],
            ["float", "Battery Bulk Charge Voltage", "V"],
            ["float", "Battery Recharge Voltage", "V"],
            ["int", "Max Charging Current", "A"],
            ["option", "Input Voltage Range", ["Appliance", "UPS"]],
            [
                "option",
                "Output Source Priority",
                ["Utility first", "Solar first", "SBU first"],
            ],
            [
                "option",
                "Charger Source Priority",
                [
                    "Utility first",
                    "Solar first",
                    "Solar + Utility",
                    "Only solar charging permitted",
                ],
            ],
            [
                "option",
                "Battery Type",
                [
                    "AGM",
                    "Flooded",
                    "User",
                    "TBD",
                    "Pylontech",
                    "WECO",
                    "Soltaro",
                    "LIb-protocol compatible",
                    "3rd party Lithium",
                ],
            ],
            ["option", "Buzzer", ["enabled", "disabled"]],
            ["option", "Power saving", ["disabled", "enabled"]],
            ["option", "Overload restart", ["disabled", "enabled"]],
            ["option", "Over temperature restart", ["disabled", "enabled"]],
            ["option", "LCD Backlight", ["disabled", "enabled"]],
            ["option", "Primary source interrupt alarm", ["disabled", "enabled"]],
            ["option", "Record fault code", ["disabled", "enabled"]],
            ["option", "Overload bypass", ["disabled", "enabled"]],
            ["option", "LCD reset to default", ["disabled", "enabled"]],
            [
                "option",
                "Output mode",
                [
                    "single machine",
                    "parallel output",
                    "Phase 1 of 3 phase output",
                    "Phase 2 of 3 phase output",
                    "Phase 3 of 3 phase output",
                    "Phase 1 of 2 phase output",
                    "Phase 2 of 2 phase output (120°)",
                    "Phase 2 of 2 phase output (180°)",
                    "Unknown Output Mode",
                ],
            ],
            ["float", "Battery Redischarge Voltage", "V"],
            [
                "option",
                "PV OK condition",
                [
                    "As long as one unit of inverters has connect PV, parallel system will consider PV OK",
                    "Only All of inverters have connect PV, parallel system will consider PV OK",
                ],
            ],
            [
                "option",
                "PV Power Balance",
                [
                    "PV input max current will be the max charged current",
                    "PV input max power will be the sum of the max charged power and loads power",
                ],
            ],
            ["int", "Max Charging Time at CV", "min"],
            ["int", "Max Discharging current", "A"],
        ],
        "test_responses": [
            b"(230.0 50.0 0030 44.0 54.0 56.4 46.0 60 0 0 2 0 0 0 0 0 1 1 1 0 1 0 54.0 0 1 224\xeb\xbc\r",
        ],
    },
    "QOPPT": {
        "name": "QOPPT",
        "description": "Device Output Source Priority Time Order Inquiry",
        "type": "QUERY",
        "response": [["bytes.decode", "Device Output Source Priority Time Order", ""]],
        "test_responses": [
            b"(2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 0 2 1>>\r",
        ],
    },
    "QCHPT": {
        "name": "QCHPT",
        "description": "Device Charger Source Priority Time Order Inquiry",
        "type": "QUERY",
        "response_type": "SEQUENTIAL",
        "response": [
            [
                "option",
                "Charger Source Priority 00 hours",
                ["undefined", "Solar first", "Solar + Utility", "Only Solar"],
            ],
            [
                "option",
                "Charger Source Priority 01 hours",
                ["undefined", "Solar first", "Solar + Utility", "Only Solar"],
            ],
            [
                "option",
                "Charger Source Priority 02 hours",
                ["undefined", "Solar first", "Solar + Utility", "Only Solar"],
            ],
            [
                "option",
                "Charger Source Priority 03 hours",
                ["undefined", "Solar first", "Solar + Utility", "Only Solar"],
            ],
            [
                "option",
                "Charger Source Priority 04 hours",
                ["undefined", "Solar first", "Solar + Utility", "Only Solar"],
            ],
            [
                "option",
                "Charger Source Priority 05 hours",
                ["undefined", "Solar first", "Solar + Utility", "Only Solar"],
            ],
            [
                "option",
                "Charger Source Priority 06 hours",
                ["undefined", "Solar first", "Solar + Utility", "Only Solar"],
            ],
            [
                "option",
                "Charger Source Priority 07 hours",
                ["undefined", "Solar first", "Solar + Utility", "Only Solar"],
            ],
            [
                "option",
                "Charger Source Priority 08 hours",
                ["undefined", "Solar first", "Solar + Utility", "Only Solar"],
            ],
            [
                "option",
                "Charger Source Priority 09 hours",
                ["undefined", "Solar first", "Solar + Utility", "Only Solar"],
            ],
            [
                "option",
                "Charger Source Priority 10 hours",
                ["undefined", "Solar first", "Solar + Utility", "Only Solar"],
            ],
            [
                "option",
                "Charger Source Priority 11 hours",
                ["undefined", "Solar first", "Solar + Utility", "Only Solar"],
            ],
            [
                "option",
                "Charger Source Priority 12 hours",
                ["undefined", "Solar first", "Solar + Utility", "Only Solar"],
            ],
            [
                "option",
                "Charger Source Priority 13 hours",
                ["undefined", "Solar first", "Solar + Utility", "Only Solar"],
            ],
            [
                "option",
                "Charger Source Priority 14 hours",
                ["undefined", "Solar first", "Solar + Utility", "Only Solar"],
            ],
            [
                "option",
                "Charger Source Priority 15 hours",
                ["undefined", "Solar first", "Solar + Utility", "Only Solar"],
            ],
            [
                "option",
                "Charger Source Priority 16 hours",
                ["undefined", "Solar first", "Solar + Utility", "Only Solar"],
            ],
            [
                "option",
                "Charger Source Priority 17 hours",
                ["undefined", "Solar first", "Solar + Utility", "Only Solar"],
            ],
            [
                "option",
                "Charger Source Priority 18 hours",
                ["undefined", "Solar first", "Solar + Utility", "Only Solar"],
            ],
            [
                "option",
                "Charger Source Priority 19 hours",
                ["undefined", "Solar first", "Solar + Utility", "Only Solar"],
            ],
            [
                "option",
                "Charger Source Priority 20 hours",
                ["undefined", "Solar first", "Solar + Utility", "Only Solar"],
            ],
            [
                "option",
                "Charger Source Priority 21 hours",
                ["undefined", "Solar first", "Solar + Utility", "Only Solar"],
            ],
            [
                "option",
                "Charger Source Priority 22 hours",
                ["undefined", "Solar first", "Solar + Utility", "Only Solar"],
            ],
            [
                "option",
                "Charger Source Priority 23 hours",
                ["undefined", "Solar first", "Solar + Utility", "Only Solar"],
            ],
            [
                "option",
                "Device Charger Source Priority",
                ["undefined", "Solar first", "Solar + Utility", "Only Solar"],
            ],
            [
                "option",
                "Selection of Charger Source Priority Order 1",
                ["undefined", "Solar first", "Solar + Utility", "Only Solar"],
            ],
            [
                "option",
                "Selection of Charger Source Priority Order 2",
                ["undefined", "Solar first", "Solar + Utility", "Only Solar"],
            ],
            [
                "option",
                "Selection of Charger Source Priority Order 3",
                ["undefined", "Solar first", "Solar + Utility", "Only Solar"],
            ],
        ],
        "test_responses": [
            b"(3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 0 0 0\xd0\x8b\r",
        ],
    },
    "QT": {
        "name": "QT",
        "description": "Device Time Inquiry",
        "type": "QUERY",
        "response_type": "SEQUENTIAL",
        "response": [["bytes.decode", "Device Time", ""]],
        "test_responses": [
            b"(20210726122606JF\r",
        ],
    },
    "QBEQI": {
        "name": "QBEQI",
        "description": "Battery Equalization Status Parameters Inquiry",
        "type": "QUERY",
        "response_type": "SEQUENTIAL",
        "response": [
            ["option", "Equalization Enabled", ["Disabled", "Enabled"]],
            ["int", "Equalization Time", "min"],
            ["int", "Equalization Period", "day"],
            ["int", "Equalization Max Current", "A"],
            ["bytes.decode", "Reserved1", ""],
            ["float", "Equalization Voltage", "V"],
            ["bytes.decode", "Reserved2", ""],
            ["int", "Equalization Over Time", "min"],
            ["option", "Equalization Active", ["Inactive", "Active"]],
            ["int", "Equalization Elasped Time", "hour"],
        ],
        "test_responses": [
            b"(1 030 030 080 021 55.40 224 030 0 0234y?\r",
        ],
    },
    "QET": {
        "name": "QET",
        "description": "Total PV Generated Energy Inquiry",
        "type": "QUERY",
        "response_type": "SEQUENTIAL",
        "response": [
            [
                "int",
                "Total PV Generated Energy",
                "Wh",
                {"icon": "mdi:solar-power", "device-class": "energy", "state_class": "total"},
            ]
        ],
        "test_responses": [
            b"(00238800!J\r",
        ],
    },
    "QEY": {
        "name": "QEY",
        "description": "Yearly PV Generated Energy Inquiry",
        "type": "QUERY",
        "response_type": "SEQUENTIAL",
        "response": [
            [
                "int",
                "PV Generated Energy for Year",
                "Wh",
                {"icon": "mdi:counter", "device-class": "energy", "state_class": "total"},
            ],
            ["info:cv", "Year", ""],
        ],
        "test_responses": [
            b"(00238800!J\r",
        ],
        "regex": "QEY(\\d\\d\\d\\d)$",
    },
    "QEM": {
        "name": "QEM",
        "description": "Monthly PV Generated Energy Inquiry",
        "type": "QUERY",
        "response_type": "SEQUENTIAL",
        "response": [
            [
                "int",
                "PV Generated Energy for Month",
                "Wh",
                {"icon": "mdi:solar-power", "device-class": "energy", "state_class": "total"},
            ],
            ["info:cv[:4]", "Year", ""],
            ["info:calendar.month_name[int(cv[4:])]", "Month", ""],
        ],
        "test_responses": [
            b"(00238800!J\r",
        ],
        "regex": "QEM(\\d\\d\\d\\d\\d\\d)$",
    },
    "QED": {
        "name": "QED",
        "description": "Daily PV Generated Energy Inquiry",
        "help": " -- display daily generated energy, format is QEDyyyymmdd",
        "type": "QUERY",
        "response_type": "INDEXED",
        "response": [
            [
                1,
                "PV Generated Energy for Day",
                "int",
                "Wh",
                {"icon": "mdi:solar-power", "device-class": "energy", "state_class": "total"},
            ],
            [2, "Year", "info:cv[:4]", ""],
            [3, "Month", "info:calendar.month_name[int(cv[4:6])]", ""],
            [4, "Day", "info:cv[6:]", ""],
        ],
        "test_responses": [
            b"(00238800!J\r",
        ],
        "regex": "QED(\\d\\d\\d\\d\\d\\d\\d\\d)$",
    },
    "QLT": {
        "name": "QLT",
        "description": "Total Output Load Energy Inquiry",
        "type": "QUERY",
        "response_type": "SEQUENTIAL",
        "response": [
            [
                "int",
                "Total Output Load Energy",
                "Wh",
                {"icon": "mdi:counter", "device-class": "energy", "state_class": "total"},
            ]
        ],
        "test_responses": [
            b"(00238800!J\r",
        ],
    },
    "QLY": {
        "name": "QLY",
        "description": "Yearly Output Load Energy Inquiry",
        "type": "QUERY",
        "response_type": "SEQUENTIAL",
        "response": [
            [
                "int",
                "Output Load Energy for Year",
                "Wh",
                {"icon": "mdi:counter", "device-class": "energy", "state_class": "total"},
            ],
            ["info:cv", "Year", ""],
        ],
        "test_responses": [
            b"(00238800!J\r",
        ],
        "regex": "QLY(\\d\\d\\d\\d)$",
    },
    "QLM": {
        "name": "QLM",
        "description": "Monthly Output Load Energy Inquiry",
        "type": "QUERY",
        "response_type": "SEQUENTIAL",
        "response": [
            [
                "int",
                "Output Load Energy for Month",
                "Wh",
                {"icon": "mdi:counter", "device-class": "energy", "state_class": "total"},
            ],
            ["info:cv[:4]", "Year", ""],
            ["info:calendar.month_name[int(cv[4:])]", "Month", ""],
        ],
        "test_responses": [
            b"(00238800!J\r",
        ],
        "regex": "QLM(\\d\\d\\d\\d\\d\\d)$",
    },
    "QLD": {
        "name": "QLD",
        "description": "Daily Output Load Energy Inquiry",
        "type": "QUERY",
        "response_type": "SEQUENTIAL",
        "response": [
            [
                "int",
                "Output Load Energy for Day",
                "Wh",
                {"icon": "mdi:counter", "device-class": "energy", "state_class": "total"},
            ],
            ["info:cv[:4]", "Year", ""],
            ["info:calendar.month_name[int(cv[4:6])]", "Month", ""],
            ["info:cv[6:]", "Day", ""],
        ],
        "test_responses": [
            b"(00238800!J\r",
        ],
        "regex": "QLD(\\d\\d\\d\\d\\d\\d\\d\\d)$",
    },
    "QLED": {
        "name": "QLED",
        "description": "LED Status Parameters Inquiry",
        "type": "QUERY",
        "response_type": "SEQUENTIAL",
        "response": [
            ["option", "LED Enabled", ["Disabled", "Enabled"]],
            ["option", "LED Speed", ["Low", "Medium", "Fast"]],
            [
                "option",
                "LED Effect",
                ["Breathing", "Unknown", "Solid", "Right Scrolling"],
            ],
            ["int", "LED Brightness", ""],
            ["int", "LED Number of Colors", ""],
            ["bytes.decode", "RGB", ""],
        ],
        "test_responses": [
            b"(1 1 2 5 3 148000211255255255000255255\xdaj\r",
        ],
    },
}

SETTER_COMMANDS = {
    "PLEDE": {
        "name": "PLEDE",
        "description": "Enable/disable LED function",
        "help": " -- examples: PLEDE0 (disable LED), PLEDE1 (enable LED)",
        "type": "SETTER",
        "response": [["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}]],
        "test_responses": [
            b"(NAK\x73\x73\r",
            b"(ACK\x39\x20\r",
        ],
        "regex": "PLEDE([01])$",
    },
    "PLEDS": {
        "name": "PLEDS",
        "description": "Set LED speed",
        "help": " -- examples: PLEDS0 (set LED speed low), PLEDS1 (set LED speed medium), PLEDS2 (set LED speed high)",
        "type": "SETTER",
        "response": [["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}]],
        "test_responses": [
            b"(NAK\x73\x73\r",
            b"(ACK\x39\x20\r",
        ],
        "regex": "PLEDS([012])$",
    },
    "PLEDM": {
        "name": "PLEDM",
        "description": "Set LED effect",
        "help": " -- examples: PLEDM0 (set LED effect breathing), PLEDM2 (set LED effect solid), PLEDM3 (set LED right scrolling)",
        "type": "SETTER",
        "response": [["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}]],
        "test_responses": [
            b"(NAK\x73\x73\r",
            b"(ACK\x39\x20\r",
        ],
        "regex": "PLEDM([0123])$",
    },
    "PLEDB": {
        "name": "PLEDB",
        "description": "Set LED brightness",
        "help": " -- examples: PLEDB1 (set LED brightness low), PLEDB5 (set LED brightness normal), PLEDB9 (set LED brightness high)",
        "type": "SETTER",
        "response": [["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}]],
        "test_responses": [
            b"(NAK\x73\x73\r",
            b"(ACK\x39\x20\r",
        ],
        "regex": "PLEDB([123456789])$",
    },
    "PLEDT": {
        "name": "PLEDT",
        "description": "Set LED total number of colors",
        "help": " -- examples: PLEDT2 (set 2 LED colors), PLEDT3 (set 3 LED colors)",
        "type": "SETTER",
        "response": [["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}]],
        "test_responses": [
            b"(NAK\x73\x73\r",
            b"(ACK\x39\x20\r",
        ],
        "regex": "PLEDT([123])$",
    },
    "PLEDC": {
        "name": "PLEDC",
        "description": "Set LED color",
        "help": " -- examples: PLEDCnRRRGGGBBB (n: 1 line mode, 2 AVR mode, 3 battery mode)",
        "type": "SETTER",
        "response": [["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}]],
        "test_responses": [
            b"(NAK\x73\x73\r",
            b"(ACK\x39\x20\r",
        ],
        "regex": "PLEDC(\\d\\d\\d\\d\\d\\d\\d\\d\\d\\d)$",
    },
}

COMMANDS_TO_REMOVE = ["QVFW2"]


class pi30max(pi30):
    def __str__(self):
        return "PI30 protocol handler for LV6048MAX and similar inverters"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__()
        self._protocol_id = b"PI30MAX"
        # Add pi30max specific commands to pi30 commands
        self.COMMANDS.update(QUERY_COMMANDS)
        # Add pi30max specific setter commands
        self.COMMANDS.update(SETTER_COMMANDS)
        # remove and unwanted pi30 commands
        for item in COMMANDS_TO_REMOVE:
            self.COMMANDS.pop(item, None)
        self.STATUS_COMMANDS = ["QPIGS", "QPIGS2"]
        self.SETTINGS_COMMANDS = ["QPIRI", "QFLAG"]
        self.DEFAULT_COMMAND = "QPI"
        log.info(f"Using protocol {self._protocol_id} with {len(self.COMMANDS)} commands")



import logging

from .abstractprotocol import AbstractProtocol

log = logging.getLogger("pi17")

QUERY_COMMANDS = {
    "PI": {
        "name": "PI",
        "prefix": "^P003",
        "description": "Device Protocol Version inquiry",
        "help": " -- queries the device protocol version",
        "type": "QUERY",
        "response": [["string", "Protocol Version", ""]],
        "test_responses": [
            b"^D00517\xca\xec\r",
        ],
    },
    "ID": {
        "name": "ID",
        "prefix": "^P003",
        "description": "Device Serial Number inquiry",
        "help": " -- queries the device serial number",
        "type": "QUERY",
        "response": [["string", "Serial Number", ""]],
        "test_responses": [
            b"^D0251496161704100242000000le\r",
        ],
    },
    "VFW": {
        "name": "VFW",
        "prefix": "^P004",
        "description": "Device CPU version inquiry",
        "help": " -- queries the CPU version",
        "type": "QUERY",
        "response": [["float", "CPU Version", ""]],
        "test_responses": [
            b"^D017VERFW:00001.01VW\r",
        ],
    },
    "VFW2": {
        "name": "VFW2",
        "prefix": "^P005",
        "description": "Device CPU 2 version inquiry",
        "help": " -- queries the CPU 2 version",
        "type": "QUERY",
        "response": [["float", "CPU 2 Version", ""]],
        "test_responses": [
            b"^D018VERFW2:00001.01\x99\xc3\r",
        ],
    },
    "MD": {
        "name": "MD",
        "prefix": "^P003",
        "description": "Device Model inquiry",
        "help": " -- queries the device model",
        "type": "QUERY",
        "response": [
            [
                "keyed",
                "Machine number",
                {
                    "000": "Infini-Solar 10KW/3P",
                    "001": "Infini-Solar 15KW/3P",
                    "002": "Infini-Solar 15KW/3P-custom",
                    "003": "Infini-Solar WP (Infini WP 12K/15K)",
                    "004": "Infini-Solar WP 30KW/3P",
                    "005": "Infini-Solar WP LV 6KW/2P",
                    "006": "Infini-Solar WP TWIN",
                },
            ],
            ["int", "Output rated VA", "kW"],
            ["int", "Output power factor", "pf"],
            ["int", "AC input phase number", "number"],
            ["int", "AC output phase number", "number"],
            ["int", "Norminal AC output voltage", "0.1V"],
            ["int", "Norminal AC input voltage", "0.1V"],
            ["string", "Battery piece number", "ea"],
            ["int", "Battery standard voltage per unit", "0.1V"],
        ],
        "test_responses": [
            b"^D037000,010000,99,3,3,2300,2300,04,120U\x82\r",
        ],
    },
    "DI": {
        "name": "DI",
        "prefix": "^P003",
        "description": "Query default value of changeable parameters",
        "help": "",
        "type": "QUERY",
        "response": [
            ["int", "AC input highest voltage for feed power", "0.1V"],
            ["int", "AC input lowest voltage for feed power", "0.1V"],
            ["int", "AC input highest frequency for feed power", "0.01Hz"],
            ["int", "AC input lowest frequency for feed power", "0.01Hz"],
            ["int", "Solar input highest MPPT voltage", "0.1V"],
            ["int", "Solar input lowest MPPT voltage", "0.1V"],
            ["int", "Solar input highest voltage", "0.1V"],
            ["int", "Solar input lowest voltage", "0.1V"],
            ["int", "Solar input highest average voltage", "0.1V"],
            ["int", "LCD sleep wait time", "30secs"],
            ["int", "Battery maximum charge current", "0.1A"],
            ["int", "Battery constant charge voltage CV", "0.1V"],
            ["int", "Battery float charge voltage", "0.1V"],
            ["int", "The wait time for feed power", "sec"],
            ["string", "Start time for support loads", "HHMM"],
            ["string", "Ending time for support loads", "HHMM"],
            ["string", "Start time for AC charger", "HHMM"],
            ["string", "Ending time for AC charger", "HHMM"],
            ["int", "Battery under voltage", "0.1V"],
            ["int", "Battery under back voltage", "0.1V"],
            ["int", "Battery weak voltage in hybrid mode", "0.1V"],
            ["int", "Battery weak back voltage in hybrid mode", "0.1V"],
            ["int", "Battery stop charge current level in floating charging", "0.1A"],
            ["int", "Keep charged time of battery catch stop charger current level", "0.1A"],
            ["int", "Battery voltage of recover to charge when battery stop charger in floating charging", "0.1V"],
        ],
        "test_responses": [],
    },
    "DM": {
        "name": "DM",
        "prefix": "^P003",
        "description": "Query machine model",
        "help": "For outputs interpretation see documentations",
        "type": "QUERY",
        "response": [["string", "model_code", ""]],
        "test_responses": [b"^D006050h\xdb\r"],
    },
    "INGS": {
        "name": "INGS",
        "prefix": "^P004",
        "description": "",
        "help": "",
        "type": "QUERY",
        "response": [
            ["int", "Input current R", "0.1A"],
            ["int", "Input current S", "0.1A"],
            ["int", "Input current T", "0.1A"],
            ["int", "Output current R", "0.1A"],
            ["int", "Output current S", "0.1A"],
            ["int", "Output current T", "0.1A"],
            ["int", "PBusVolt", "0.1V"],
            ["int", "NBusVolt", "0.1V"],
            ["int", "PBusAvgV", "0.1V"],
            ["int", "NBusAvgV", "0.1V"],
            ["int", "NLintCur", "0.1A"],
        ],
        "test_responses": [b"^D0560020,0019,0021,0002,0004,0005,3809,3809,3810,3807,000\xf1\x1e\r"],
    },
    "EMINFO": {
        "name": "EMINFO",
        "prefix": "^P005",
        "description": "",
        "help": "",
        "type": "QUERY",
        "response": [
            ["int", "EMFirst", ""],
            ["int", "DefFeed-InPow", ""],
            ["int", "ActPvPow", ""],
            ["int", "ActFeedPow", ""],
            ["int", "ReservPow", ""],
            ["int", "EMLast", ""],
        ],
        "test_responses": [b"^D0301,10000,00005,00010,00000,1\xad\xc4\r"],
    },
    "PIRI": {
        "name": "PIRI",
        "prefix": "^P005",
        "description": "Device rated information",
        "help": " -- queries rated information",
        "type": "QUERY",
        "response": [
            ["int", "AC input rated voltage", "0.1V"],
            ["string", "AC input rated frequency", "Hz"],
            ["string", "AC input rated current", "0.1A"],
            ["string", "AC output rated voltage", "0.1V"],
            ["string", "AC output rated current", "0.1A"],
            ["string", "MPPT rated current per string", "0.1A"],
            ["string", "Battery rated voltage", "0.1V"],
            ["string", "MPPT track number", "ea"],
            [
                "keyed",
                "Machine type",
                {
                    "00": "Grid type",
                    "01": "Off-grid type",
                    "10": "Hybrid type",
                },
            ],
            ["option", "Topology", ["transformerless", "transformer"]],
            ["option", "Parallel for output", ["disabled", "enabled"]],
        ],
        "test_responses": [
            b"^D0452400,500,0416,2400,0416,0187,0480,2,10,0,1\xbcs\r",
        ],
    },
    "GS": {
        "name": "GS",
        "prefix": "^P003",
        "description": "Query general status",
        "help": " -- queries general status",
        "type": "QUERY",
        "response": [
            ["int", "Solar input voltage 1", "0.1V"],
            ["int", "Solar input voltage 2", "0.1V"],
            ["int", "Solar input current 1", "0.1A"],
            ["int", "Solar input current 2", "0.1A"],
            ["int", "Battery voltage", "0.1V"],
            ["int", "Battery capacity", "%"],
            ["int", "Battery current", "0.1A"],
            ["int", "AC input voltage R", "0.1V"],
            ["int", "AC input voltage S", "0.1V"],
            ["int", "AC input voltage T", "0.1V"],
            ["int", "AC input frequency", "0.01Hz"],
            ["int", "AC input current R", "0.1A"],
            ["int", "AC input current S", "0.1A"],
            ["int", "AC input current T", "0.1A"],
            ["int", "AC output voltage R", "0.1V"],
            ["int", "AC output voltage S", "0.1V"],
            ["int", "AC output voltage T", "0.1V"],
            ["int", "AC output frequency", "0.01Hz"],
            ["int", "AC output current R", "0.1A"],
            ["int", "AC output current S", "0.1A"],
            ["int", "AC output current T", "0.1A"],
            ["int", "Inner temperature", "°C"],
            ["int", "Component max temperature", "°C"],
            ["int", "External Battery temperature", "°C"],
            [
                "option",
                "Setting change bit",
                ["No setting change", "Settings changed - please refresh"],
            ],
        ],
        "test_responses": [
            b"^D1100000,0000,0000,0000,0394,000,+00000,2389,2427,2459,5002,0000,0000,0000,2378,2434,2455,5001,,,,029,029,000,0\xf8n\r",
        ],
    },
    "PS": {
        "name": "PS",
        "prefix": "^P003",
        "description": "Device Power Status",
        "help": " -- queries power status",
        "type": "QUERY",
        "response": [
            ["int", "Solar input power 1", "W"],
            ["int", "Solar input power 2", "W"],
            ["int", "Battery power", "W"],
            ["int", "AC input active power R", "W"],
            ["int", "AC input active power S", "W"],
            ["int", "AC input active power T", "W"],
            ["int", "AC input total active power", "W"],
            ["int", "AC output active power R", "W"],
            ["int", "AC output active power S", "W"],
            ["int", "AC output active power T", "W"],
            ["int", "AC output total active power", "W"],
            ["int", "AC output apparent power R", "VA"],
            ["int", "AC output apparent power S", "VA"],
            ["int", "AC output apparent power T", "VA"],
            ["int", "AC output total apparent power", "VA"],
            ["int", "AC output power percentage", "%"],
            ["option", "AC output connect status", ["Disconnected", "Connected"]],
            ["option", "Solar input 1 work status", ["Idle", "Working"]],
            ["option", "Solar input 2 work status", ["Idle", "Working"]],
            ["option", "Battery power direction", ["Idle", "Charging", "Discharging"]],
            ["option", "DC/AC power direction", ["Idle", "AC to DC", "DC to AC"]],
            ["option", "Line power direction", ["Idle", "Input", "Output"]],
        ],
        "test_responses": [
        ],
    },
    "MOD": {
        "name": "MOD",
        "prefix": "^P004",
        "description": "Device working mode inquiry",
        "help": " -- queries the device working mode",
        "type": "QUERY",
        "response": [
            [
                "keyed",
                "Working mode",
                {
                    "00": "Power on mode",
                    "01": "Standby mode",
                    "02": "Bypass mode",
                    "03": "Battery mode",
                    "04": "Fault mode",
                    "05": "Hybrid mode (Line mode, Grid mode)",
                    "06": "Charge mode",
                },
            ],
        ],
        "test_responses": [
            b"^D00505\xd9\x9f\r",
        ],
    },
    "WS": {
        "name": "WS",
        "prefix": "^P003",
        "description": "Warning status inquiry",
        "help": " -- queries any active warnings flags from the Inverter",
        "type": "QUERY",
        "response": [
            ["option", "Solar input 1 loss", ["disabled", "enabled"]],
            ["option", "Solar input 2 loss", ["disabled", "enabled"]],
            ["option", "Solar input 1 voltage too high", ["disabled", "enabled"]],
            ["option", "Solar input 2 voltage too high", ["disabled", "enabled"]],
            ["option", "Battery under voltage", ["disabled", "enabled"]],
            ["option", "Battery low voltage", ["disabled", "enabled"]],
            ["option", "Battery disconnected", ["disabled", "enabled"]],
            ["option", "Battery over voltage", ["disabled", "enabled"]],
            ["option", "Battery low in hybrid mode", ["disabled", "enabled"]],
            ["option", "Grid voltage high loss", ["disabled", "enabled"]],
            ["option", "Grid voltage low loss", ["disabled", "enabled"]],
            ["option", "Grid frequency high loss", ["disabled", "enabled"]],
            ["option", "Grid frequency low loss", ["disabled", "enabled"]],
            ["option", "AC input long-time average voltage over", ["disabled", "enabled"]],
            ["option", "AC input voltage loss", ["disabled", "enabled"]],
            ["option", "AC input frequency loss", ["disabled", "enabled"]],
            ["option", "AC input island", ["disabled", "enabled"]],
            ["option", "AC input phase dislocation", ["disabled", "enabled"]],
            ["option", "Over temperature", ["disabled", "enabled"]],
            ["option", "Over load", ["disabled", "enabled"]],
            ["option", "Emergency Power Off active", ["disabled", "enabled"]],
            ["option", "AC input wave loss", ["disabled", "enabled"]],
        ],
        "test_responses": [
            b"^D0471,1,0,0,1,1,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,\x14\x9c\r",
        ],
    },
    "FLAG": {
        "name": "FLAG",
        "prefix": "^P005",
        "description": "Query enable/disable flag status",
        "help": " -- queries enable/disable flag status from the Inverter",
        "type": "QUERY",
        "response": [
            ["option", "Mute buzzer beep", ["Disabled", "Enabled"]],
            ["option", "Mute buzzer beep in standby mode", ["Disabled", "Enabled"]],
            [
                "option",
                "Mute buzzer beep only on battery discharged status",
                ["Disabled", "Enabled"],
            ],
            ["option", "Generator as AC input", ["Disabled", "Enabled"]],
            ["option", "Wide AC input range", ["Disabled", "Enabled"]],
            ["option", "N/G relay function", ["Disabled", "Enabled"]],
        ],
        "test_responses": [
            b"^D0120,0,1,0,1\xd8\xf2\r",
        ],
    },
    "T": {
        "name": "T",
        "prefix": "^P002",
        "description": "Query current time",
        "help": " -- queries current time from the Inverter",
        "type": "QUERY",
        "response": [
            ["string", "DateTime", "YYYYMMDDHHMMSS"],
        ],
        "test_responses": [
            b"^D01720210521234743\x0eR\r",
        ],
    },
    "ET": {
        "name": "ET",
        "prefix": "^P003",
        "description": "Query total generated energy",
        "help": " -- queries total generated energy from the Inverter",
        "type": "QUERY",
        "response": [
            [
                "int",
                "Generated energy",
                "kWh",
                {"icon": "mdi:counter", "device-class": "energy", "state_class": "total"},
            ],
        ],
        "test_responses": [
            b"^D01100006591\xba\x10\r",
        ],
    },
    "BATS": {
        "name": "BATS",
        "description": "Query battery setting",
        "help": " -- queries battery setting",
        "type": "QUERY",
        "response": [
            ["int", "Battery maximum charge current", "0.1A"],
            ["int", "Battery constant charge voltage(C.V.)", "0.1V"],
            ["int", "Battery floating charge voltage", "0.1V"],
            ["int", "Battery stop charger current level in floating charging", "0.1A"],
            [
                "int",
                "Keep charged time of battery catch stopped charging current level",
                "Minutes",
            ],
            [
                "int",
                "Battery voltage of recover to charge when battery stop charger in floating charging",
                "0.1V",
            ],
            ["int", "Battery under voltage", "0.1V"],
            ["int", "Battery under voltage release", "0.1V"],
            ["int", "Battery weak voltage in hybrid mode", "0.1V"],
            ["int", "Battery weak voltage release in hybrid mode", "0.1V"],
            ["option", "Battery Type", ["Ordinary", "Li-Fe"]],
            ["string", "Reserved", ""],
            ["string", "Battery install date", "YYYYMMDDHHMMSS"],
            [
                "option",
                "AC charger keep battery voltage function enable/diable",
                ["Disabled", "Enabled"],
            ],
            ["int", "AC charger keep battery voltage", "0.1V"],
            ["int", "Battery temperature sensor compensation", "0.m1V"],
            ["int", "Max. AC charging current", "0.1A"],
            ["int", "Battery discharge max current in hybrid mode", "A"],
            ["option", "Enable/Disable EPS function", ["Disabled", "Enabled"]],
            ["int", "Battery voltage of cut-off Main output in battery mode(", "0.1V"],
            [
                "int",
                "Battery voltage of re-connecting Main output in battery mode",
                "0.1V",
            ],
        ],
        "test_responses": [
            b"^D0762000,0584,0576,0000,000,0576,0460,0510,0460,0510,1,,,1,0540,000,2000,0250\x85Y\r",
            b"^D0941750,0560,0540,0000,060,0530,0420,0480,0480,0540,0,,,0,0480,000,0100,0175,000,000,000,000,0\xc9\xd9\r",
            b"^D0941750,0560,0540,0000,060,0530,0420,0480,0480,0540,0,,,0,0480,000,0100,0175,010,020,020,080,0mr\r",
        ],
    },
    "HECS": {
        "name": "HECS",
        "prefix": "^P005",
        "description": "Query energy control status",
        "help": " -- queries the device energy distribution",
        "type": "QUERY",
        "response": [
            [
                "keyed",
                "Solar Energy Distribution Priority",
                {
                    "00": "Battery-Load-Grid",
                    "01": "Load-Battery-Grid",
                    "02": "Load-Grid-Battery",
                },
            ],
            ["option", "Solar charge battery", ["disabled", "enabled"]],
            ["option", "AC charge battery", ["disabled", "enabled"]],
            ["option", "Feed power to utility", ["disabled", "enabled"]],
            ["option", "Battery discharge to loads when solar input normal", ["disabled", "enabled"]],
            ["option", "Battery discharge to loads when solar input loss", ["disabled", "enabled"]],
            ["option", "Battery discharge to feed grid when solar input normal", ["disabled", "enabled"]],
            ["option", "Battery discharge to feed grid when solar input loss", ["disabled", "enabled"]],
            ["option", "Reserved", ["disabled", "enabled"]],
        ],
        "test_responses": [
            b"^D01900,0,0,0,0,0,0,0,0\r",
        ],
    },
    "EY": {
        "name": "EY",
        "prefix": "^P010",
        "description": "Query generated energy of year",
        "help": " -- queries generated energy for the year YYYY from the Inverter",
        "type": "QUERYEN",
        "response": [
            ["int", "Generated energy", "Wh"],
        ],
        "test_responses": [
            b"^D01100006591\xba\x10\r",
        ],
        "regex": "EY(\\d\\d\\d\\d)$",
    },
    "EM": {
        "name": "EM",
        "prefix": "^P012",
        "description": "Query generated energy of month",
        "help": " -- queries generated energy for the month YYYYMM from the Inverter",
        "type": "QUERYEN",
        "response": [
            ["int", "Generated energy", "Wh"],
        ],
        "test_responses": [
            b"^D01000006591\xba\x10\r",
        ],
        "regex": "EM(\\d\\d\\d\\d\\d\\d)$",
    },
    "ED": {
        "name": "ED",
        "prefix": "^P014",
        "description": "Query generated energy of day",
        "help": " -- queries generated energy for the day YYYYMMDD from the Inverter",
        "type": "QUERYEN",
        "response": [
            ["int", "Generated energy", "Wh"],
        ],
        "test_responses": [
            b"^D009000091\xba\x10\r",
        ],
        "regex": "ED(\\d\\d\\d\\d\\d\\d\\d\\d)$",
    },
    "EH": {
        "name": "EH",
        "prefix": "^P016",
        "description": "Query generated energy of hour",
        "help": " -- queries generated energy for the hour YYYYMMDDHH from the Inverter",
        "type": "QUERYEN",
        "response": [
            ["int", "Generated energy", "Wh"],
        ],
        "test_responses": [
            b"^D008000001\xba\x10\r",
        ],
        "regex": "EH(\\d\\d\\d\\d\\d\\d\\d\\d\\d\\d)$",
    },
}

SETTER_COMMANDS = {
    "LON": {
        "name": "LON",
        "description": "Set enable/disable machine supply power to the loads",
        "help": " -- examples: LON1 (0: disable, 1: enable)",
        "type": "SETTER",
        "response": [
            ["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}],
        ],
        "test_responses": [
            b"^1\x0b\xc2\r",
            b"^0\x1b\xe3\r",
        ],
        "regex": "LON([01])$",
    },
    "PA": {
        "name": "PA",
        "description": "Mute buzzer beep",
        "help": " -- examples: PA1 (1: enable, 0: disable)",
        "type": "SETTER",
        "response": [
            ["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}],
        ],
        "test_responses": [
            b"^1\x0b\xc2\r",
            b"^0\x1b\xe3\r",
        ],
        "regex": "PA([01])$",
    },
    "PB": {
        "name": "PB",
        "description": "Mute buzzer beep in standby mode",
        "help": " -- examples: PB1 (1: enable, 0: disable)",
        "type": "SETTER",
        "response": [
            ["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}],
        ],
        "test_responses": [
            b"^1\x0b\xc2\r",
            b"^0\x1b\xe3\r",
        ],
        "regex": "PB([01])$",
    },
    "PC": {
        "name": "PC",
        "description": "Mute buzzer beep only on battery discharged status",
        "help": " -- examples: PC1 (1: enable, 0: disable)",
        "type": "SETTER",
        "response": [
            ["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}],
        ],
        "test_responses": [
            b"^1\x0b\xc2\r",
            b"^0\x1b\xe3\r",
        ],
        "regex": "PC([01])$",
    },
    "PD": {
        "name": "PD",
        "description": "Generator as AC input",
        "help": " -- examples: PD1 (1: enable, 0: disable)",
        "type": "SETTER",
        "response": [
            ["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}],
        ],
        "test_responses": [
            b"^1\x0b\xc2\r",
            b"^0\x1b\xe3\r",
        ],
        "regex": "PD([01])$",
    },
    "PE": {
        "name": "PE",
        "description": "Wide AC input range",
        "help": " -- examples: PE1 (1: enable, 0: disable)",
        "type": "SETTER",
        "response": [
            ["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}],
        ],
        "test_responses": [
            b"^1\x0b\xc2\r",
            b"^0\x1b\xe3\r",
        ],
        "regex": "PE([01])$",
    },
    "PF": {
        "name": "PF",
        "description": "N/G relay function",
        "help": " -- examples: PF1 (1: enable, 0: disable)",
        "type": "SETTER",
        "response": [
            ["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}],
        ],
        "test_responses": [
            b"^1\x0b\xc2\r",
            b"^0\x1b\xe3\r",
        ],
        "regex": "PF([01])$",
    },
    "DAT": {
        "name": "DAT",
        "description": "Set date time",
        "help": " -- examples: DAT190518224530(YYMMDDHHMMSS-12digits)",
        "type": "SETTER",
        "response": [
            ["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}],
        ],
        "test_responses": [
            b"^1\x0b\xc2\r",
            b"^0\x1b\xe3\r",
        ],
        "regex": "DAT(\\d\\d\\d\\d\\d\\d\\d\\d\\d\\d\\d\\d)$",
    },
    "LST": {
        "name": "LST",
        "description": "Set LCD sleep wait time",
        "help": " -- examples: LSTnn (nn: 00, 01, 02, 10, 20 for selection, unit : 30second.00 means LCD always light)",
        "type": "SETTER",
        "response": [
            ["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}],
        ],
        "test_responses": [
            b"^1\x0b\xc2\r",
            b"^0\x1b\xe3\r",
        ],
        "regex": "LST((0[123])|([12]0))$",
    },
    "EDA": {
        "name": "EDA",
        "description": "Enable/disable solar charge battery",
        "help": " -- examples: EDA1 (1: enable, 0: disable)",
        "type": "SETTER",
        "response": [
            ["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}],
        ],
        "test_responses": [
            b"^1\x0b\xc2\r",
            b"^0\x1b\xe3\r",
        ],
        "regex": "EDA([01])$",
    },
    "EDB": {
        "name": "EDB",
        "description": "Enable/disable AC charge battery",
        "help": " -- examples: EDB1 (1: enable, 0: disable)",
        "type": "SETTER",
        "response": [
            ["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}],
        ],
        "test_responses": [
            b"^1\x0b\xc2\r",
            b"^0\x1b\xe3\r",
        ],
        "regex": "EDB([01])$",
    },
    "EDC": {
        "name": "EDC",
        "description": "Enable/disable feed power to utility",
        "help": " -- examples: EDC1 (1: enable, 0: disable)",
        "type": "SETTER",
        "response": [
            ["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}],
        ],
        "test_responses": [
            b"^1\x0b\xc2\r",
            b"^0\x1b\xe3\r",
        ],
        "regex": "EDC([01])$",
    },
    "EDD": {
        "name": "EDD",
        "description": "Enable/disable battery discharge to loads when solar input normal",
        "help": " -- examples: EDD1 (1: enable, 0: disable)",
        "type": "SETTER",
        "response": [
            ["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}],
        ],
        "test_responses": [
            b"^1\x0b\xc2\r",
            b"^0\x1b\xe3\r",
        ],
        "regex": "EDD([01])$",
    },
    "EDE": {
        "name": "EDE",
        "description": "Enable/disable battery discharge to loads when solar input loss",
        "help": " -- examples: EDE1 (1: enable, 0: disable)",
        "type": "SETTER",
        "response": [
            ["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}],
        ],
        "test_responses": [
            b"^1\x0b\xc2\r",
            b"^0\x1b\xe3\r",
        ],
        "regex": "EDE([01])$",
    },
    "EDF": {
        "name": "EDF",
        "description": "Enable/disable battery discharge to feed power to utility when solar input normal",
        "help": " -- examples: EDF1 (1: enable, 0: disable)",
        "type": "SETTER",
        "response": [
            ["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}],
        ],
        "test_responses": [
            b"^1\x0b\xc2\r",
            b"^0\x1b\xe3\r",
        ],
        "regex": "EDF([01])$",
    },
    "EDG": {
        "name": "EDG",
        "description": "Enable/disable battery discharge to feed power to utility whensolar input loss",
        "help": " -- examples: EDG1 (1: enable, 0: disable)",
        "type": "SETTER",
        "response": [
            ["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}],
        ],
        "test_responses": [
            b"^1\x0b\xc2\r",
            b"^0\x1b\xe3\r",
        ],
        "regex": "EDG([01])$",
    },
    "BT": {
        "name": "BT",
        "description": "Set battery type",
        "help": " -- examples: EDG1 (1: Li-Fe, 0: Ordinary)",
        "type": "SETTER",
        "response": [
            ["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}],
        ],
        "test_responses": [
            b"^1\x0b\xc2\r",
            b"^0\x1b\xe3\r",
        ],
        "regex": "BT([01])$",
    },
    "MCHGV": {
        "name": "MCHGV",
        "description": "Set battery charge voltages",
        "help": " -- examples: MCHGV0576,0566 (CV voltage in 0.1V xxxx, Float voltage xxxx in 0.1V)",
        "type": "SETTER",
        "response": [
            ["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}],
        ],
        "test_responses": [
            b"^1\x0b\xc2\r",
            b"^0\x1b\xe3\r",
        ],
        "regex": "MCHGV(0[45]\\d\\d,0[45]\\d\\d)$",
    },
    "ACCT": {
        "name": "ACCT",
        "description": "Set AC charge time range",
        "help": " -- examples: ACCT2200-0259 (Sets time range from 22:00 to 02:59. End minute is inclusive)",
        "type": "SETTER",
        "response": [
            ["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}],
        ],
        "test_responses": [
            b"^1\x0b\xc2\r",
            b"^0\x1b\xe3\r",
        ],
        "regex": "ACCT((2[0-3]|[01]?[0-9])([0-5]?[0-9])-(2[0-3]|[01]?[0-9])([0-5]?[0-9]))$",
    },
    "ACCB": {
        "name": "ACCB",
        "description": "AC Charger  keep  battery voltage",
        "help": " -- examples: ACCB1,0450 (1: enable, 0: disable),(400-600) 0.1V",
        "type": "SETTER",
        "response": [
            ["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}],
        ],
        "test_responses": [
            b"^1\x0b\xc2\r",
            b"^0\x1b\xe3\r",
        ],
        "regex": "ACCB([01],0[456]\\d\\d)$",
    },
    "MCHGC": {
        "name": "MCHGC",
        "description": "Set battery maximum charge current",
        "help": " -- examples: MCHGC1200 (Current in mA xxxx)",
        "type": "SETTER",
        "response": [
            ["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}],
        ],
        "test_responses": [
            b"^1\x0b\xc2\r",
            b"^0\x1b\xe3\r",
        ],
        "regex": "MCHGC([012]\\d\\d\\d)$",
    },
    "MUCHGC": {
        "name": "MUCHGC",
        "description": "Set maximum charge current from AC",
        "help": " -- examples: MUCHGC0600 (Current in mA xxxx)",
        "type": "SETTER",
        "response": [
            ["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}],
        ],
        "test_responses": [
            b"^1\x0b\xc2\r",
            b"^0\x1b\xe3\r",
        ],
        "regex": "MUCHGC([012]\\d\\d\\d)$",
    },
    "SEP": {
        "name": "SEP",
        "description": "Set solar energy distribution priority",
        "help": " -- examples: SEP(00-BLG; 01-LBG; 02-LGB)",
        "type": "SETTER",
        "response": [
            ["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}],
        ],
        "test_responses": [
            b"^1\x0b\xc2\r",
            b"^0\x1b\xe3\r",
        ],
        "regex": "SEP(0[012])$",
    },
    "BDCM": {
        "name": "BDCM",
        "description": "Battery discharge max current in hybrid mode",
        "help": " -- examples: (BDCMxxxx, 0.1A)",
        "type": "SETTER",
        "response": [
            ["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}],
        ],
        "test_responses": [
            b"^1\x0b\xc2\r",
            b"^0\x1b\xe3\r",
        ],
        "regex": "BDCM(0[1-2]\\d\\d)$",
    },
    "BATDV": {
        "name": "BATDV",
        "description": "Battery discharge voltage limits",
        "help": " -- examples: (BATDV0420,0440,0430,0450, 0.1V)",
        "type": "SETTER",
        "response": [
            ["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}],
        ],
        "test_responses": [
            b"^1\x0b\xc2\r",
            b"^0\x1b\xe3\r",
        ],
        "regex": "BATDV(0[45]\\d\\d,0[45]\\d\\d,0[45]\\d\\d,0[45]\\d\\d)$",
    },
}


class pi17(AbstractProtocol):
    def __str__(self):
        return "PI17 protocol handler"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__()
        self._protocol_id = b"PI17"
        self.COMMANDS = QUERY_COMMANDS
        self.COMMANDS.update(SETTER_COMMANDS)
        # TODO fix these lists
        self.STATUS_COMMANDS = []
        self.SETTINGS_COMMANDS = [
            "MD",
        ]
        self.DEFAULT_COMMAND = "PI"
        self.PID = "PI"
        self.ID_COMMANDS = ["PI", "DM"]

    def get_full_command(self, command) -> bytes:
        """
        Override the default get_full_command as its different
        """
        log.info(f"Using protocol {self._protocol_id} with {len(self.COMMANDS)} commands")
        # These need to be set to allow other functions to work`
        self._command = command
        self._command_defn = self.get_command_defn(command)
        # End of required variables setting
        if self._command_defn is None:
            return None

        _cmd = bytes(self._command, "utf-8")
        _type = self._command_defn["type"]
        # No CRC in PI17 commands?
        data_length = len(_cmd) + 1
        if _type == "QUERY":
            _prefix = f"^P{data_length:03}"
            _pre_cmd = bytes(_prefix, "utf-8") + _cmd
            log.debug(f"_pre_cmd: {_pre_cmd}")
            log.debug(f"_prefix: {_prefix}")
            # calculate the CRC
            # crc_high; crc_low = crc(_pre_cmd)
            # combine byte_cmd, CRC , return
            # PI18 full command "^P005GS\x..\x..\r"
            # _crc = bytes([crc_high, crc_low, 13])
            full_command = _pre_cmd + bytes([13])  # + _crc
            log.debug(f"full command: {full_command}")
            return full_command
        elif _type == "QUERYEN":
            data_length1 = len(_cmd) + 4
            _prefix = f"^P{data_length1:03}"
            log.debug(f"_prefix: {_prefix}")
            intermedstr = _prefix + self._command
            _numb0 = sum(bytearray(intermedstr, "utf-8")) & 255
            _numb = f"{_numb0:03d}"
            log.debug(f"_numb: {_numb}")
            _pre_cmd = intermedstr + str(_numb)
            log.debug(f"_pre_cmd: {_pre_cmd}")
            full_command = bytes(_pre_cmd, "utf-8") + bytes([13])
            log.debug(f"full command: {full_command}")
            return full_command
        else:
            _prefix = f"^S{data_length:03}"
            _pre_cmd = bytes(_prefix, "utf-8") + _cmd
            log.debug(f"_pre_cmd: {_pre_cmd}")
            # calculate the CRC
            # crc_high; crc_low = crc(_pre_cmd)
            # combine byte_cmd, CRC , return
            # PI18 full command "^P005GS\x..\x..\r"
            # _crc = bytes([crc_high, crc_low, 13])
            full_command = _pre_cmd + bytes([13])  # + _crc
            log.debug(f"full command: {full_command}")
            return full_command

    def get_responses(self, response):
        """
        Override the default get_responses as its different
        """
        responses = response.split(b",")
        if responses[0] == b"^0\x1b\xe3\r":
            # is a reject response
            return ["NAK"]
        elif responses[0] == b"^1\x0b\xc2\r":
            # is a successful acknowledgement response
            return ["ACK"]

        # Drop ^Dxxx from first response
        responses[0] = responses[0][5:]
        # Remove CRC of last response
        responses[-1] = responses[-1][:-3]
        return responses



""" result.py """
import logging
from enum import Enum, auto

from powermon.commands.reading import Reading
from powermon.commands.reading_definition import ReadingDefinition, ResponseType
from powermon.dto.resultDTO import ResultDTO

log = logging.getLogger("result")


class ResultType(Enum):
    """ enum of valid types of Results """
    ERROR = auto()
    ACK = auto()  # ack / nak type result, normally from a setter command
    SINGLE = auto()  # single value in result
    ORDERED = auto()  # the order of the values determines what they are
    SLICED = auto()  # the response needs to be sliced into separate values
    MULTIVALUED = auto()  # the response has multiple values, but they all correspond to one result
    VED_INDEXED = auto()  # the response has a key / value pair (separated by \t, each pair separated by \r\n), with the key used to find the definition


class Result:
    """
    object to contain all the info of a result, including
    - command definition <- should this be the command object???
    - 'raw response' from the device
    - list of Readings (processed results)
    """
    def __str__(self):
        return f"Result: {self.is_valid=}, {self.error=} - {self.error_messages=}, {self.raw_response=}, " + ','.join(str(reading) for reading in self._readings)

    # def __init__(self, result_type: ResultType, command_definition, raw_response: bytes, responses: list | dict):
    def __init__(self, command, raw_response: bytes, responses: list | dict):
        self.is_valid = True
        self.error = False
        self.error_messages = []

        self.command = command
        self.result_type = command.command_definition.result_type
        # self.command_definition = command.command_definition

        self.raw_response = raw_response
        self.readings: list[Reading] = responses

        log.debug("Result: %s", self)

    @property
    def raw_response(self):
        """ the raw_response as received from the device """
        return self._raw_response

    @raw_response.setter
    def raw_response(self, value):
        if value is None:
            raise ValueError("raw_response cannot be None")
        self._raw_response = value

    @property
    def result_type(self):
        """ the type of the result """
        return self._result_type

    @result_type.setter
    def result_type(self, value):
        self._result_type = value
        if value == ResultType.ERROR:
            self.is_valid = False
            self.error = True
            self.error_messages = [self.raw_response]

    @property
    def readings(self) -> list[Reading]:
        """ list of processed readings """
        return self._readings

    @readings.setter
    def readings(self, responses):
        self._readings = self.decode_responses(responses=responses)

    def add_readings(self, readings: list[Reading]) -> bool:
        """ add a list of readings to the current list """
        self._readings.extend(readings)
        return True

    def decode_responses(self, responses=None) -> list[Reading]:
        """
        Take the response and decode into a list of Readings depending on the result type
        """
        log.info("result.response passed to decode: %s, result_type %s", responses, self.result_type)
        all_readings : list[Reading] = []

        if responses is None:
            return all_readings

        # Process response based on result type
        match self.result_type:
            case ResultType.ACK | ResultType.SINGLE | ResultType.MULTIVALUED:
                # Get the reading definition (there is only one)
                reading_definition: ReadingDefinition = self.command.command_definition.get_reading_definition()
                # Process the response using the reading_definition, into readings
                readings = self.readings_from_response(responses, reading_definition)
                all_readings.extend(readings)
            case ResultType.ORDERED | ResultType.SLICED:
                # Have a list of reading_definitions and a list of responses that correspond to each other
                # possibly additional INFO definitions (at end of definition list??)
                definition_count = self.command.command_definition.reading_definition_count()
                response_count = len(responses)
                for position in range(definition_count):
                    reading_definition: ReadingDefinition = self.command.command_definition.get_reading_definition(position=position)
                    if position < response_count:
                        readings = self.readings_from_response(responses[position], reading_definition)
                        all_readings.extend(readings)
                    else:
                        # More definitions than results, either INFO type definitions or too little data
                        if reading_definition.response_type == ResponseType.INFO_FROM_COMMAND:
                            # INFO is contained in supplied command eg QEY2023 -> 2023
                            readings = self.readings_from_response(self.command.code, reading_definition)
                            all_readings.extend(readings)
            case ResultType.VED_INDEXED:
                # have a list of (index,value) tuples
                for key, value in responses:
                    reading_definition: ReadingDefinition = self.command.command_definition.get_reading_definition(lookup=key)
                    if reading_definition is not None:
                        # Process the response using the reading_definition, into readings
                        readings = self.readings_from_response(value, reading_definition)
                        all_readings.extend(readings)
            case ResultType.ERROR:
                # Get the reading_definition - this may need fixing for errors
                reading_definition: ReadingDefinition = self.command.command_definition.get_reading_definition()
                # Process the response using the reading_definition, into readings
                readings = self.readings_from_response(responses, reading_definition)
                all_readings.extend(readings)
            case _:
                # unknown result type
                raise ValueError(f"Unknown result type: {self.result_type}")
        log.debug("got readings: %s", ",".join(str(i) for i in all_readings))
        # print(','.join(str(reading) for reading in all_readings))
        return all_readings

    def readings_from_response(self, response, reading_definition) -> list[Reading]:
        """ return readings from a raw_response using the supplied reading definition """
        try:
            return reading_definition.reading_from_raw_response(response, override=self.command.override)
        except ValueError:
            error = Reading(data_name=reading_definition.description,
                            data_value=reading_definition.get_invalid_message(response), data_unit="")
            error.is_valid = False
            return [error]

    def to_dto(self) -> ResultDTO:
        """ convert result object to data transfer object """
        reading_dtos = []
        for reading in self.readings:
            reading_dtos.append(reading.to_dto())
        return ResultDTO(device_identifier="self.device_id", command_code="self.command_code", data=reading_dtos)


class ResultError(Result):
    """ docstring todo """
    def __init__(self, command, raw_response: bytes, responses: list | dict):
        self.command = command
        self.raw_response = raw_response
        self.result_type = ResultType.ERROR

        self.is_valid = False
        self.error = True
        self.error_messages = responses

        # self.command_definition = command.command_definition

        self.readings: list[Reading] = None

        log.debug("Result: %s", self)



from strenum import LowercaseStrEnum
from enum import auto
import datetime

class ParameterType(LowercaseStrEnum):
    DATE = auto()
    
class ParameterFormat(LowercaseStrEnum):
    YYYYMMDD = auto()
    
class Parameter:
    def __init__(self, name, description, parameter_type: ParameterType, parameter_format: ParameterFormat):
        self.name = name
        self.description = description
        self.parameter_type = parameter_type
        self.parameter_format = parameter_format
        self.value = None
        
    def set_value(self, value):
        if self.parameter_type == ParameterType.DATE:
            if self.parameter_format == ParameterFormat.YYYYMMDD:
                self.value = datetime.date(int(value[0:4]), int(value[4:6]), int(value[6:8]))
        
        
    @classmethod
    def multiple_from_config(cls, parameters_config: list) -> dict[str, "Parameter"]:
        parameters: dict[str, Parameter] = {}
        if parameters_config is None:
            return parameters
        for parameter_config in parameters_config:
            parameter = Parameter.from_config(parameter_config)
            parameters[parameter.name] = parameter
        return parameters
        
    @classmethod
    def from_config(cls, parameter_config: dict) -> "Parameter":
        name = parameter_config.get("name")
        description = parameter_config.get("description")
        parameter_type = ParameterType(parameter_config.get("parameter_type"))
        parameter_format = ParameterFormat(parameter_config.get("parameter_format"))
        return Parameter(name, description, parameter_type, parameter_format)
        
    def __str__(self):
        return f"Parameter: {self.name} {self.description} {self.parameter_type} {self.parameter_format}"
        
    def __repr__(self):
        return self.__str__()
        


# import importlib
import logging
from enum import Enum, auto

from mppsolar.helpers import get_kwargs


class PortType(Enum):
    UNKNOWN = auto()
    TEST = auto()
    USB = auto()
    ESP32 = auto()
    SERIAL = auto()
    JKBLE = auto()
    MQTT = auto()
    VSERIAL = auto()
    DALYSERIAL = auto()
    REMOTESOCKET = auto()


log = logging.getLogger("io")


def get_port_type(port):
    if port is None:
        return PortType.UNKNOWN

    port = port.lower()
    # check for test type port
    if "test" in port:
        log.debug("port matches test")
        return PortType.TEST
    # mqtt
    elif "mqtt" in port:
        log.debug("port matches mqtt")
        return PortType.MQTT
    # USB type ports
    elif "hidraw" in port:
        log.debug("port matches hidraw")
        return PortType.USB
    elif "mppsolar" in port:
        log.debug("port matches mppsolar")
        return PortType.USB
    # ESP type ports
    elif "esp" in port:
        log.debug("port matches esp")
        return PortType.ESP32
    # JKBLE type ports
    elif ":" in port:
        # all mac addresses currently return as JKBLE
        log.debug("port matches jkble ':'")
        return PortType.JKBLE
    elif "jkble" in port:
        log.debug("port matches jkble")
        return PortType.JKBLE
    elif "daly" in port:
        log.debug("port matches daly")
        return PortType.DALYSERIAL
    elif "remotesocket" in port:
        log.debug("port matches remotesocket")
        return PortType.REMOTESOCKET
    elif "vserial" in port:
        log.debug("port matches vserial")
        return PortType.VSERIAL
    elif "serial" in port:
        log.debug("port matches serial")
        return PortType.SERIAL
    elif "ttyusb" in port:
        log.debug("port matches ttyusb")
        return PortType.SERIAL
    else:
        return PortType.UNKNOWN


def get_port(*args, **kwargs):
    port = get_kwargs(kwargs, "port")
    baud = get_kwargs(kwargs, "baud", 2400)
    porttype = get_kwargs(kwargs, "porttype", None)

    if porttype:
        log.info(f"Port overide - using port '{porttype}'")
        port_type = get_port_type(porttype)
    else:
        port_type = get_port_type(port)

    if port_type == PortType.TEST:
        log.info("Using testio for communications")
        from mppsolar.inout.testio import TestIO

        _port = TestIO(device_path=port)

    elif port_type == PortType.USB:
        log.info("Using hidrawio for communications")
        from mppsolar.inout.hidrawio import HIDRawIO

        _port = HIDRawIO(device_path=port)

    elif port_type == PortType.ESP32:
        log.info("Using esp32io for communications")
        from mppsolar.inout.esp32io import ESP32IO

        _port = ESP32IO(device_path=port)

    elif port_type == PortType.JKBLE:
        log.info("Using jkbleio for communications")
        from mppsolar.inout.jkbleio import JkBleIO

        _port = JkBleIO(device_path=port)

    elif port_type == PortType.SERIAL:
        log.info("Using serialio for communications")
        from mppsolar.inout.serialio import SerialIO

        _port = SerialIO(device_path=port, serial_baud=baud)

    elif port_type == PortType.DALYSERIAL:
        log.info("Using dalyserialio for communications")
        from mppsolar.inout.dalyserialio import DalySerialIO

        _port = DalySerialIO(device_path=port, serial_baud=baud)

    elif port_type == PortType.VSERIAL:
        log.info("Using vserialio for communications")
        from mppsolar.inout.vserialio import VSerialIO

        _port = VSerialIO(device_path=port, serial_baud=baud, records=30)

    elif port_type == PortType.REMOTESOCKET:
        log.info("Using remotesocketio for communications")
        if ":" not in port:
            log.error(
                "port must be in format 'ip_address:port' for remote socket connection"
            )
            return None
        remote_ip, remote_port = port.split(":")
        log.debug(f"got ip: {remote_ip}, port: {remote_port}")
        from mppsolar.inout.remotesocketio import remoteSocketIO

        _port = remoteSocketIO(remote_ip=remote_ip, remote_port=remote_port)

    elif port_type == PortType.MQTT:
        mqtt_broker = get_kwargs(kwargs, "mqtt_broker", "localhost")
        name = get_kwargs(kwargs, "name", "unnamed")
        # mqtt_port = get_kwargs(kwargs, "mqtt_port", 1883)
        # mqtt_user = get_kwargs(kwargs, "mqtt_user")
        # mqtt_pass = get_kwargs(kwargs, "mqtt_pass")
        log.info(f"Using mqttio for communications broker {mqtt_broker}")

        from mppsolar.inout.mqttio import MqttIO

        _port = MqttIO(
            client_id=name,
            mqtt_broker=mqtt_broker,
            # mqtt_port=mqtt_port,
            # mqtt_user=mqtt_user,
            # mqtt_pass=mqtt_pass,
        )

    else:
        _port = None
    return _port



""" errors.py - collection of powermon specific exceptions """


class ConfigError(Exception):
    """ Exception for invaild configurations """


class PowermonProtocolError(Exception):
    """ Exception for errors with protocol definitions """


class PowermonWIP(Exception):
    """ Exception for work not yet done """


class InvalidResponse(Exception):
    """ Exception for invalid responses """


class InvalidCRC(Exception):
    """ Exception for invalid crc """


class CommandDefinitionMissing(Exception):
    """ Exception for missing / not found command definition """


class CommandError(Exception):
    """ Exception for errors with commands """



""" commands / command.py """
import logging

from powermon.commands.command_definition import CommandDefinition
from powermon.commands.result import Result, ResultType, ResultError
from powermon.commands.trigger import Trigger
from powermon.dto.commandDTO import CommandDTO
from powermon.errors import ConfigError, InvalidResponse, InvalidCRC
from powermon.outputs import OutputType, multiple_from_config
from powermon.outputs.abstractoutput import AbstractOutput
from powermon.outputs.api_mqtt import ApiMqtt

log = logging.getLogger("Command")


class Command():
    """
    Command object, holds the details of the command, including:
    - command code
    - command type
    - command definition (inc overrides)
    - trigger
    - outputs
    """
    def __str__(self):
        if self.code is None:
            return "empty command object"

        last_run = self.trigger.get_last_run()
        next_run = self.trigger.get_next_run()

        _outs = ""
        for output in self.outputs:
            _outs += str(output)

        return f"Command: {self.code=} {self.full_command=}, {self.type=}, \
            [{_outs=}], {last_run=}, {next_run=}, {str(self.trigger)}, {str(self.command_definition)} {self.override=}"

    def __init__(self, code: str, commandtype: str, outputs: list[AbstractOutput], trigger: Trigger):
        self.code = code
        self.type = commandtype

        self.outputs = outputs
        self.trigger: Trigger = trigger

        self.full_command = None

    @classmethod
    def from_config(cls, config=None) -> "Command":
        """build object from config dict"""
        # need to have a config defined
        # minimum is
        # - command: QPI
        if not config:
            log.warning("Invalid command config")
            raise ConfigError("Invalid command config")
            # return None

        code = config.get("command")
        if code is None:
            log.info("command must be defined")
            raise ConfigError("command must be defined in config")
        commandtype = config.get("type", "basic")
        override = config.get("override", None)
        # if override is not None:
        #     print("override: %s" % override)
        outputs = multiple_from_config(config.get("outputs", ""))
        trigger = Trigger.from_config(config=config.get("trigger"))
        command_object = cls(code=code, commandtype=commandtype, outputs=outputs, trigger=trigger)
        command_object.override = override
        return command_object

    @classmethod
    def from_dto(cls, command_dto: CommandDTO) -> "Command":
        """ build object from data transfer object """
        trigger = Trigger.from_dto(command_dto.trigger)
        command = cls(code=command_dto.command_code, commandtype="basic", outputs=[], trigger=trigger)
        outputs = []
        for output_dto in command_dto.outputs:
            if output_dto.type == OutputType.API_MQTT:
                outputs.append(ApiMqtt.from_dto(output_dto))
        command.outputs = outputs
        return command

    def build_result(self, raw_response=None, protocol=None) -> Result:
        """ build a result object from the raw_response """
        log.debug("build_result: for command with 'code: %s, command_definition: %s'", self.code, self.command_definition)
        try:
            # check response is valid
            protocol.check_valid(raw_response, self.command_definition)
            # check crc is correct
            protocol.check_crc(raw_response, self.command_definition)
        except (InvalidResponse, InvalidCRC) as e:
            _result = ResultError(command=self, raw_response=e, responses=[e.__context__, str(e)])
            _result.result_type = ResultType.ERROR
            # print(_result)
            return _result

        # trim response
        trimmed_response = protocol.trim_response(raw_response, self.command_definition)
        # split response
        responses = protocol.split_response(trimmed_response, self.command_definition)
        # build the Result object
        result = Result(command=self, raw_response=raw_response, responses=responses)
        return result

    @property
    def full_command(self) -> str | None:
        """return the full command, including CRC and/or headers"""
        return self._full_command

    @full_command.setter
    def full_command(self, full_command):
        """ store the full command """
        self._full_command = full_command

    @property
    def command_definition(self) -> CommandDefinition:
        """ the definition of this command """
        return self._command_definition

    @command_definition.setter
    def command_definition(self, command_definition: CommandDefinition):
        """store the definition of the command"""
        log.debug("Setting command_definition to: %s", command_definition)
        if command_definition is None:
            raise ValueError("CommandDefinition cannot be None")

        # Check if the definition is valid for the command
        if command_definition.is_command_code_valid(self.code) is False:
            raise ValueError(f"Command code {self.code} is not valid for command definition regex {command_definition.regex}")
        self._command_definition = command_definition

    @property
    def override(self):
        """ dict of override options """
        # use getattr and return None if _override not set
        return getattr(self, "_override", None)

    @override.setter
    def override(self, value):
        self._override = value

    @property
    def outputs(self) -> list[AbstractOutput]:
        """ a list of output objects """
        return self._outputs

    @outputs.setter
    def outputs(self, outputs: list[AbstractOutput]):
        self._outputs = outputs

    def is_due(self):
        """ is this command due to run? """
        return self.trigger.is_due()

    def touch(self):
        """ update trigger run time """
        self.trigger.touch()

    def to_dto(self):
        """ return the command data transfer object """
        return CommandDTO(
            command_code=self.code,
            device_id="not set",
            result_topic=self.outputs[0].get_topic(),
            trigger=self.trigger.to_dto(),
            outputs=[output.to_dto() for output in self.outputs],
        )



""" mppsolar / inout / jkbleio.py """
import logging

try:
    from bluepy import btle
except ImportError:
    print("You are missing dependencies in order to be able to use that output.")
    print("To install them, use that command:")
    print("    python -m pip install 'mppsolar[ble]'")

from .baseio import BaseIO
from ..helpers import get_kwargs
from .jkbledelegate import jkBleDelegate

log = logging.getLogger("JkBleIO")

getInfo = (
    b"\xaa\x55\x90\xeb\x97\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x11"
)
# getInfo = b"\xaa\x55\x90\xeb\x97\x00\xdf\x52\x88\x67\x9d\x0a\x09\x6b\x9a\xf6\x70\x9a\x17\xfd"


class JkBleIO(BaseIO):
    def __init__(self, device_path) -> None:
        self._device = None
        self._device_path = device_path
        self.maxConnectionAttempts = 3
        self.record = None

    def send_and_receive(self, *args, **kwargs) -> dict:
        # Send the full command via the communications port
        command = get_kwargs(kwargs, "command")
        protocol = get_kwargs(kwargs, "protocol")
        full_command = protocol.get_full_command(command)
        log.info(f"full command {full_command} for command {command}")

        command_defn = protocol.get_command_defn(command)
        record_type = command_defn["record_type"]
        log.debug(f"expected record type {record_type} for command {command}")

        # Connect to BLE device
        if self.ble_connect(self._device_path, protocol, record_type):
            response = self.ble_get_data(full_command)
            self.ble_disconnect()
        else:
            log.error(f"Failed to connect to {self._device_path}")
            response = None
        # End of BLE device connection
        log.debug(f"Raw response {response}")
        return response

    def ble_connect(self, mac=None, protocol=None, record_type=0x02):
        """
        Connect to a BLE device with 'mac' address
        """
        self._device = None
        # Intialise BLE device
        self._device = btle.Peripheral(None)
        self._device.withDelegate(jkBleDelegate(self, protocol, record_type))
        # Connect to BLE Device
        connected = False
        attempts = 0
        log.info(f"Attempting to connect to {mac}")
        while not connected:
            attempts += 1
            if attempts > self.maxConnectionAttempts:
                log.warning(
                    f"Cannot connect to mac {mac} - exceeded {attempts - 1} attempts"
                )
                return connected
            try:
                self._device.connect(mac)
                self._device.setMTU(330)
                connected = True
            except Exception:
                continue
        return connected

    def ble_disconnect(self):
        log.info("Disconnecting BLE Device...")
        self._device.disconnect()
        self._device = None
        return

    def ble_get_data(self, command=None):
        self.record = None

        log.debug(f"Command: {command}")

        if command is None:
            return self.record

        # Get the device name
        try:
            serviceId = self._device.getServiceByUUID(btle.AssignedNumbers.genericAccess)
            deviceName = serviceId.getCharacteristics(btle.AssignedNumbers.deviceName)[0]
            log.info("Connected to {}".format(deviceName.read()))
        except btle.BTLEGattError as e:
            log.warning(f"Error getting device name: {e}")

        # Connect to the notify service
        serviceNotifyUuid = "ffe0"
        serviceNotify = self._device.getServiceByUUID(serviceNotifyUuid)

        # Get the handles that we need to talk to
        # Read
        # characteristicReadUuid = "ffe3"  #old version
        characteristicReadUuid = (
            "ffe1"  # TODO: need to validate this works for older bms
        )
        characteristicRead = serviceNotify.getCharacteristics(characteristicReadUuid)[0]
        handleRead = characteristicRead.getHandle()
        log.info(
            "Read characteristic: {}, handle {:x}".format(
                characteristicRead, handleRead
            )
        )

        # ## TODO sort below
        # Need to dynamically find this handle....
        log.info(
            "Enable 0x0b handle %s", self._device.writeCharacteristic(0x0B, b"\x01\x00")
        )
        log.info(
            "Enable read handle %s",
            self._device.writeCharacteristic(handleRead, b"\x01\x00")
        )
        log.info(
            "Write getInfo to read handle %s",
            self._device.writeCharacteristic(handleRead, getInfo)
        )
        secs = 0
        while True:
            if self._device.waitForNotifications(1.0):
                continue
            secs += 1
            if secs > 5:
                break

        if command == "getInfo":
            return self.record[:300]

        log.info(
            "Write command to read handle",
            self._device.writeCharacteristic(handleRead, command),
        )
        loops = 0
        recordsToGrab = 1
        log.info("Grabbing {} records (after inital response)".format(recordsToGrab))

        while True:
            loops += 1
            if loops > recordsToGrab * 15 + 16:
                log.info("jkbleio: ble_get_dataa: Got {} records".format(recordsToGrab))
                break
            if self._device.waitForNotifications(1.0):
                continue

        log.debug(f"Record now {self.record} len {len(self.record)}")
        return self.record[:300]



import logging

from .baseoutput import baseoutput
from ..helpers import get_kwargs

log = logging.getLogger("raw")


class raw(baseoutput):
    def __str__(self):
        return "outputs the raw results to standard out"

    def __init__(self, *args, **kwargs) -> None:
        log.debug(f"__init__: kwargs {kwargs}")

    def output(self, *args, **kwargs):
        log.info("Using output processor: raw")
        log.debug(f"kwargs {kwargs}")
        data = get_kwargs(kwargs, "data")
        if data is None:
            return
        _desc = data.pop("_command_description", "No description found")
        if command := data.pop("_command", None):
            print(f"Command: {command} - {_desc}")
            print("-" * 60)
        if "raw_response" in data:
            key = "raw_response"
            value = data[key][0]
            print(f"{key:<30}\t{value!a:<15}")
        return



""" formats / __init__.py """
import logging
from enum import auto

from strenum import LowercaseStrEnum

from powermon.formats.abstractformat import AbstractFormat

log = logging.getLogger("formats")


class FormatterType(LowercaseStrEnum):
    """ enumeration of valid formatter types """
    HASS = auto()
    HTMLTABLE = auto()
    RAW = auto()
    SIMPLE = auto()
    JSON = auto()
    TABLE = auto()
    TOPICS = auto()


DEFAULT_FORMAT = FormatterType.SIMPLE


def getFormatfromConfig(formatConfig) -> AbstractFormat:
    # Get values from config
    log.debug("getFormatfromConfig, formatConfig: %s", formatConfig)

    # formatConfig can be None, a str (eg 'simple') or a dict
    if formatConfig is None:
        formatType = FormatterType.SIMPLE
        formatConfig = {}
    elif isinstance(formatConfig, str):
        formatType = formatConfig
        formatConfig = {}
        formatConfig["type"] = formatType
    else:
        formatType = formatConfig.get("type")
    log.debug("getFormatfromConfig, formatType: %s", formatType)

    formatter = None
    # TODO: should we replace this config processing with from_config methods on each type to remain consistent?
    match formatType:
        case FormatterType.HTMLTABLE:
            from powermon.formats.htmltable import htmltable
            formatter = htmltable(formatConfig)
        case FormatterType.HASS:
            from powermon.formats.hass import Hass
            formatter = Hass(formatConfig)
        # case FormatterType.TOPICS:
        #     from powermon.formats.topics import Topics
        #     formatter = Topics(formatConfig)
        case FormatterType.SIMPLE:
            from powermon.formats.simple import SimpleFormat
            formatter = SimpleFormat(formatConfig)
        case FormatterType.TABLE:
            from powermon.formats.table import Table
            formatter = Table(formatConfig)
        case FormatterType.RAW:
            from powermon.formats.raw import raw
            formatter = raw(formatConfig)
        case _:
            log.warning("No formatter found for: %s", formatType)
            formatter = None

    return formatter



""" reading_definition.py """
import calendar  # pylint: disable=w0611 # needed for INFO type evaluating templates
import logging
from enum import auto
from struct import unpack

from strenum import LowercaseStrEnum

from powermon.commands.reading import Reading

log = logging.getLogger("ReadingDefinition")


class ResponseType(LowercaseStrEnum):
    """
    the type of the response
    - determines how to read and translate to useful info
    """
    ACK = auto()
    BOOL = auto()      # 0 is false, 1 is true
    INT = auto()
    HEX_CHAR = auto()
    FLOAT = auto()
    STRING = auto()
    BYTES = auto()
    LE_2B_S = auto()  # little endian 2 byte value (as string, eg 7800 = 0x0078 = 120)
    OPTION = auto()  # response identifies which option from a dict (called 'options') is the info
    LIST = auto()  # response identifies which option from a list (called 'options') is the info
    BIT_ENCODED = auto()
    ENABLE_DISABLE_FLAGS = auto()
    FLAGS = auto()
    INFO = auto()
    INFO_FROM_COMMAND = auto()  # process the supplied command using a template
    TEMPLATE_BYTES = auto()  # process the response value using a template
    TEMPLATE_INT = auto()  # process the response int value using a template


class ReadingType(LowercaseStrEnum):
    """
    the type of the reading
    - higher level type, like Wh etc
    - allows translations
    """
    ACK = auto()
    CURRENT = auto()
    APPARENT_POWER = auto()
    ENERGY = auto()
    WATTS = auto()
    WATT_HOURS = auto()
    KILOWATT_HOURS = auto()
    VOLTS = auto()
    DATE_TIME = auto()
    YEAR = auto()
    MONTH = auto()
    DAY = auto()
    TIME = auto()
    TIME_SECONDS = auto()
    TIME_MINUTES = auto()
    TIME_HOURS = auto()
    TIME_DAYS = auto()
    MESSAGE = auto()
    MESSAGE_AMPS = auto()
    FLAGS = auto()
    MULTI_ENABLE_DISABLE = auto()
    TEMPERATURE = auto()
    PERCENTAGE = auto()
    FREQUENCY = auto()
    HEX_STR = auto()


class ReadingDefinition():
    """
    Default / base ReadingDefinition
    It doesn't contain the response value, just the definition of what is valid.
    """
    def __str__(self):
        return f"{self.index=}, {self.description=}, {self.response_type=}, {self.unit=}, instance={type(self)}"

    def __init__(self, index, response_type, description, device_class, state_class, icon, unit=""):
        # {"index": 13, "reading_type": ReadingType.WATTS, "response_type": ResponseType.INT,
        #  "description": "SCC charge power", "icon": "mdi:solar-power", "device-class": "power"}
        self.description = description
        self.response_type = response_type
        self.unit = unit
        self.index = index

        self.device_class = device_class
        self.state_class = state_class
        self.icon = icon

    @property
    def description(self) -> str:
        """ text description of this reading """
        return self._description

    @description.setter
    def description(self, value):
        """ set the description """
        # log.debug("Setting description to '%s'", value)
        self._description = value

    @property
    def response_type(self) -> ResponseType:
        """ response_type of this reading """
        return self._response_type

    @response_type.setter
    def response_type(self, value):
        # log.debug("Setting response_type to '%s'", value)
        self._response_type = value

    @property
    def options(self) -> dict:
        """ options dict for decoding """
        return self._options

    @options.setter
    def options(self, value):
        self._options = value

    @property
    def slice_array(self) -> str:
        """ slice array for decoding """
        return self._slice_array

    @slice_array.setter
    def slice_array(self, value):
        self._slice_array = value

    @property
    def default(self):
        """ default value for error situations """
        return self._default

    @default.setter
    def default(self, value):
        self._default = value

    @property
    def format_template(self):
        """ a template that allows re-formating of value """
        return self._format_template

    @format_template.setter
    def format_template(self, value):
        if value is not None:
            log.debug("format templater for '%s': %s", self.description, value)
        self._format_template = value

    def translate_raw_response(self, raw_value):
        """ interpret the raw response into a python basic type """
        log.debug("translate_raw_response: %s from type: %s", raw_value, self.response_type)
        match self.response_type:
            case ResponseType.BOOL:
                return bool(int(raw_value.decode('utf-8')))
            case ResponseType.HEX_CHAR:
                return raw_value[0]
                # return ord(raw_value.decode('utf-8')[0])
            case ResponseType.INT:
                try:
                    result = int(raw_value.decode('utf-8'))
                    return result
                except ValueError as e:
                    if self.default:
                        return self.default
                    raise ValueError(f"For Reading Defininition '{self.description}', expected an INT, got {raw_value}") from e
            case ResponseType.TEMPLATE_INT:
                try:
                    r = int(raw_value.decode('utf-8'))
                    if self.format_template:
                        r = eval(self.format_template)  # pylint: disable=W0123
                    return r
                except ValueError as e:
                    if self.default:
                        return self.default
                    raise ValueError(f"For Reading Defininition '{self.description}', expected an INT, got {raw_value}") from e
            case ResponseType.FLOAT:
                return float(raw_value.decode('utf-8'))
            case ResponseType.LE_2B_S:
                result = raw_value.decode('utf-8')
                result = unpack('<h', bytes.fromhex(result))[0]
                return result
            case ResponseType.BIT_ENCODED:
                if not isinstance(self.options, dict):
                    raise TypeError(f"For Reading Defininition '{self.description}', options must be a dict if response_type is BIT_ENCODED")
                value = int(raw_value.decode('utf-8'))
                if value == 0:
                    return self.options[0]
                # loop through options and check if applicable
                results = []
                for key in self.options.keys():
                    if value & key:
                        results.append(self.options[key])
                return ",".join(results)
            case ResponseType.OPTION:
                if not isinstance(self.options, dict):
                    raise TypeError(f"For Reading Defininition '{self.description}', options must be a dict if response_type is OPTION")
                value = str(raw_value.decode('utf-8'))
                try:
                    return self.options[value]
                except KeyError as e:
                    raise KeyError(f"For Reading Defininition '{self.description}', keys: {self.options.keys()}, requested key: {value}") from e
            case ResponseType.LIST:
                if not isinstance(self.options, list):
                    raise TypeError(f"For Reading Defininition '{self.description}', options must be a list if response_type is LIST")
                value = int(raw_value.decode('utf-8'))
                try:
                    return self.options[value]
                except IndexError as e:
                    raise IndexError(f"For Reading Defininition '{self.description}', len: {len(self.options)}, got index: {value}") from e
            case ResponseType.TEMPLATE_BYTES:
                r = raw_value.decode('utf-8')
                if self.format_template:
                    r = eval(self.format_template)  # pylint: disable=W0123
                return r
            case ResponseType.INFO_FROM_COMMAND:
                cn = raw_value
                if self.format_template:
                    # print(self.format_template)
                    res = eval(self.format_template)  # pylint: disable=W0123
                    return res
                return cn
            case _:
                return raw_value.decode('utf-8')

    def reading_from_raw_response(self, raw_value, override=None) -> list[Reading]:
        """ generate a reading object from a raw value """
        log.debug("raw_value: %s, override: %s", raw_value, override)
        value = self.translate_raw_response(raw_value)
        return [
            Reading(
                data_name=self.description,
                data_value=value,
                data_unit=self.unit,
                device_class=self.device_class,
                state_class=self.state_class,
                icon=self.icon,
            )
        ]

    def get_invalid_message(self, raw_value) -> str:
        """ message for invalid state """
        return f"Invalid response for {self.description}: {raw_value}"

    def is_info(self) -> bool:
        """ is this reading definition an info definition """
        return False

    @classmethod
    def multiple_from_config(cls, reading_definition_configs: list[dict]) -> dict[int | str, "ReadingDefinition"]:
        """ build list of reading definitions from config """
        if reading_definition_configs is None:
            return {}
        else:
            reading_definitions: dict[int, "ReadingDefinition"] = {}
            for i, reading_definition_config in enumerate(reading_definition_configs):
                if "index" in reading_definition_config:
                    i = reading_definition_config.get("index")
                reading_definition = cls.from_config(reading_definition_config, i)
                log.debug("reading definition: %s", reading_definition)
                reading_definitions[reading_definition.index] = reading_definition
            return reading_definitions

    @classmethod
    def from_config(cls, reading_definition_config: dict, i) -> "ReadingDefinition":
        """ build a reading definition object from a config dict """
        index = i
        description = reading_definition_config.get("description")
        response_type = reading_definition_config.get("response_type")
        reading_type = reading_definition_config.get("reading_type")
        device_class = reading_definition_config.get("device_class", None)
        state_class = reading_definition_config.get("state_class", None)
        icon = reading_definition_config.get("icon", None)

        match reading_type:
            case ReadingType.ACK:
                reading = ReadingDefinitionACK(
                    index=index, response_type=response_type, description=description,
                    device_class=device_class, state_class=state_class, icon=icon)
            case ReadingType.WATT_HOURS:
                reading = ReadingDefinitionNumeric(
                    index=index, response_type=response_type, description=description,
                    device_class=device_class, state_class=state_class, icon=icon)
                reading.unit = "Wh"
            case ReadingType.KILOWATT_HOURS:
                reading = ReadingDefinitionNumeric(
                    index=index, response_type=response_type, description=description,
                    device_class=device_class, state_class=state_class, icon=icon)
                reading.unit = "kWh"
            case ReadingType.APPARENT_POWER:
                reading = ReadingDefinitionNumeric(
                    index=index, response_type=response_type, description=description,
                    device_class=device_class, state_class=state_class, icon=icon)
                reading.unit = "VA"
            case ReadingType.ENERGY:
                reading = ReadingDefinitionNumeric(
                    index=index, response_type=response_type, description=description,
                    device_class=device_class, state_class=state_class, icon=icon)
                reading.unit = "Ah"
            case ReadingType.MESSAGE:
                reading = ReadingDefinitionMessage(
                    index=index, response_type=response_type, description=description,
                    device_class=device_class, state_class=state_class, icon=icon)
            case ReadingType.MESSAGE_AMPS:
                reading = ReadingDefinitionMessage(
                    index=index, response_type=response_type, description=description,
                    device_class=device_class, state_class=state_class, icon=icon)
                reading.unit = "A"
            case ReadingType.TEMPERATURE:
                reading = ReadingDefinitionTemperature(
                    index=index, response_type=response_type, description=description,
                    device_class=device_class, state_class=state_class, icon=icon)
            case ReadingType.MULTI_ENABLE_DISABLE:
                reading = ReadingDefinitionENFlags(
                    index=index, response_type=response_type, description=description,
                    device_class=device_class, state_class=state_class, icon=icon)
            case ReadingType.DATE_TIME:
                reading = ReadingDefinitionMessage(
                    index=index, response_type=response_type, description=description,
                    device_class=device_class, state_class=state_class, icon=icon)
            case ReadingType.YEAR:
                reading = ReadingDefinitionMessage(
                    index=index, response_type=response_type, description=description,
                    device_class=device_class, state_class=state_class, icon=icon)
            case ReadingType.MONTH:
                reading = ReadingDefinitionMessage(
                    index=index, response_type=response_type, description=description,
                    device_class=device_class, state_class=state_class, icon=icon)
            case ReadingType.DAY:
                reading = ReadingDefinitionMessage(
                    index=index, response_type=response_type, description=description,
                    device_class=device_class, state_class=state_class, icon=icon)
            case ReadingType.TIME_SECONDS:
                reading = ReadingDefinitionNumeric(
                    index=index, response_type=response_type, description=description,
                    device_class=device_class, state_class=state_class, icon=icon)
                reading.unit = "sec"
            case ReadingType.TIME_MINUTES:
                reading = ReadingDefinitionNumeric(
                    index=index, response_type=response_type, description=description,
                    device_class=device_class, state_class=state_class, icon=icon)
                reading.unit = "min"
            case ReadingType.TIME_HOURS:
                reading = ReadingDefinitionNumeric(
                    index=index, response_type=response_type, description=description,
                    device_class=device_class, state_class=state_class, icon=icon)
                reading.unit = "hours"
            case ReadingType.TIME_DAYS:
                reading = ReadingDefinitionNumeric(
                    index=index, response_type=response_type, description=description,
                    device_class=device_class, state_class=state_class, icon=icon)
                reading.unit = "days"
            case ReadingType.FLAGS:
                flags = reading_definition_config.get("flags")
                reading = ReadingDefinitionFlags(
                    index=index, response_type=response_type, description=description, flags=flags,
                    device_class=device_class, state_class=state_class, icon=icon)
            case ReadingType.CURRENT:
                reading = ReadingDefinitionNumeric(
                    index=index, response_type=response_type, description=description,
                    device_class=device_class, state_class=state_class, icon=icon)
                reading.unit = "A"
            case ReadingType.VOLTS:
                reading = ReadingDefinitionNumeric(
                    index=index, response_type=response_type, description=description,
                    device_class=device_class, state_class=state_class, icon=icon)
                reading.unit = "V"
            case ReadingType.PERCENTAGE:
                reading = ReadingDefinitionNumeric(
                    index=index, response_type=response_type, description=description,
                    device_class=device_class, state_class=state_class, icon=icon)
                reading.unit = "%"
            case ReadingType.FREQUENCY:
                reading = ReadingDefinitionNumeric(
                    index=index, response_type=response_type, description=description,
                    device_class=device_class, state_class=state_class, icon=icon)
                reading.unit = "Hz"
            case ReadingType.WATTS:
                reading = ReadingDefinitionNumeric(
                    index=index, response_type=response_type, description=description,
                    device_class=device_class, state_class=state_class, icon=icon)
                reading.unit = "W"
            case ReadingType.HEX_STR:
                reading = ReadingDefinitionHexStr(
                    index=index, response_type=response_type, description=description,
                    device_class=device_class, state_class=state_class, icon=icon)
            case _:
                log.error("Reading description: %s has unknown reading_type definition type: %s", description, reading_type)
                raise ValueError(
                    f"Reading description: {description} has unknown reading_type definition type: {reading_type}"
                )
        # Use options dict to supply additional decode data
        # currently in use for ResponseType.OPTIONS and ENABLE_DISABLE_FLAGS
        options: dict[str, str] = reading_definition_config.get("options")
        if options is not None:
            reading.options = options
        # check for a default setting
        reading.default = reading_definition_config.get("default")
        # check for a format template
        reading.format_template = reading_definition_config.get("format_template")
        # check for a slice value
        reading.slice_array = reading_definition_config.get("slice")
        return reading


class ReadingDefinitionNumeric(ReadingDefinition):
    """ A ReadingDefinition for readings that must be numeric """
    def __init__(self, index: int, response_type: str, description: str, device_class: str = None, state_class: str = None, icon: str = None):
        super().__init__(index, response_type, description, device_class, state_class, icon)
        if response_type not in [ResponseType.INT, ResponseType.TEMPLATE_INT, ResponseType.FLOAT, ResponseType.LE_2B_S, ResponseType.HEX_CHAR]:
            raise TypeError(f"{type(self)} response must be of type int or float, ResponseType {response_type} is not valid")


class ReadingDefinitionACK(ReadingDefinition):
    """ ReadingDefinition for ACK type readings """
    def __init__(self, index: int, response_type: str, description: str, device_class: str = None, state_class: str = None, icon: str = None, ):
        super().__init__(index, response_type, description, device_class, state_class, icon)

        self.fail_code = "NAK"
        self.fail_description = "Failed"
        self.success_code = "ACK"
        self.success_description = "Succeeded"

    def reading_from_raw_response(self, raw_value, override=None) -> list[Reading]:
        value = raw_value.decode()
        if value == self.success_code:
            return [
                Reading(data_name=self.description, data_value=self.success_description, data_unit=None,
                        device_class=self.device_class, state_class=self.state_class, icon=self.icon)
            ]
        elif value == self.fail_code:
            return [
                Reading(data_name=self.description, data_value=self.fail_description, data_unit=None,
                        device_class=self.device_class, state_class=self.state_class, icon=self.icon)
            ]


class ReadingDefinitionMessage(ReadingDefinition):
    """ ReadingDefinition for message (ie wordy) type readings """
    def __init__(self, index: int, response_type: str, description: str, device_class: str = None, state_class: str = None, icon: str = None):
        super().__init__(index, response_type, description, device_class, state_class, icon)


class ReadingDefinitionHexStr(ReadingDefinitionNumeric):
    """ ReadingDefinition for hex (displayed as a string) type readings """
    def reading_from_raw_response(self, raw_value, override=None) -> list[Reading]:
        """ generate a reading object from a raw value """
        log.debug("raw_value: %s, override: %s", raw_value, override)
        value = self.translate_raw_response(raw_value)
        value = hex(value)
        return [
            Reading(
                data_name=self.description,
                data_value=value,
                data_unit=self.unit,
                device_class=self.device_class,
                state_class=self.state_class,
                icon=self.icon,
            )
        ]


class ReadingDefinitionTemperature(ReadingDefinitionNumeric):
    """ ReadingDefinition for temperature readings that are stored in celcius - can be overriden to translate to fahrenheit """
    def __init__(self, index: int, description: str, response_type: ResponseType, device_class: str = None, state_class: str = None, icon: str = None):
        super().__init__(index, response_type, description, device_class, state_class, icon)
        self.unit = "°C"

    def reading_from_raw_response(self, raw_value, override=None) -> list[Reading]:
        """ generate a reading object from a raw value """
        log.debug("raw_value: %s, override: %s", raw_value, override)
        value = self.translate_raw_response(raw_value)
        _unit = self.unit
        if override is not None and "temperature" in override:
            temp_override = override.get('temperature')
            if temp_override.startswith("F"):
                value = (1.8 * value) + 32
                _unit = "°F"
        return [
            Reading(
                data_name=self.description,
                data_value=value,
                data_unit=_unit,
                device_class=self.device_class,
                state_class=self.state_class,
                icon=self.icon,
            )
        ]


class ReadingDefinitionENFlags(ReadingDefinition):
    """ ReadingDefinition for specific Enable/Disable flag (eg: EakxyDbjuvz) type readings """
    def __init__(self, index: int, description: str, response_type: ResponseType, device_class: str = None, state_class: str = None, icon: str = None):
        super().__init__(index, response_type, description, device_class, state_class, icon)

    def translate_raw_response(self, raw_value) -> dict[str, str]:
        return_values = {}
        status = "unknown"
        for i, item in enumerate(raw_value):
            item = chr(item)
            if item == "E":
                status = "enabled"
            elif item == "D":
                status = "disabled"
            else:
                _key = self.options.get(item, {}) or f"unknown_{i}"
                return_values[_key] = status
        return return_values

    def reading_from_raw_response(self, raw_value, override=None) -> list[Reading]:
        values: dict[str, str] = self.translate_raw_response(raw_value)
        responses = []
        for name, value in values.items():
            responses.append(
                Reading(data_name=name, data_value=value, data_unit="", device_class=self.device_class, state_class=self.state_class, icon=self.icon)
            )

        return responses


class ReadingDefinitionFlags(ReadingDefinition):
    """ ReadingDefinition for flags (eg: 10100110) type readings """
    def __init__(self, index: int, description: str, response_type: ResponseType, flags: list[str], device_class: str = None, state_class: str = None, icon: str = None):
        super().__init__(index, response_type, description, device_class, state_class, icon)
        self.flags = flags

    def translate_raw_response(self, raw_value) -> dict[str, int]:
        return_value = {}
        for i, flag in enumerate(raw_value):
            if self.flags[i]:  # only append value if flag name is present
                return_value[self.flags[i]] = int(chr(flag))
        return return_value

    def reading_from_raw_response(self, raw_value, override=None) -> list[Reading]:
        values: dict[str, int] = self.translate_raw_response(raw_value)
        responses = []
        for name, value in values.items():
            responses.append(
                Reading(
                    data_name=name, data_value=value, data_unit="bool", device_class=self.device_class, state_class=self.state_class, icon=self.icon
                )
            )
        return responses



import logging
from powermon.formats.abstractformat import AbstractFormat
from powermon.commands.result import Result
from powermon.commands.reading import Reading

log = logging.getLogger("htmltable")


class htmltable(AbstractFormat):
    def __init__(self, formatConfig):
        super().__init__(formatConfig)
        self.name = "htmltable"
        
    # def set_command_description(self, command_description):
    #     pass

    def format(self, result: Result, device_info):
        log.info("Using output formatter: %s", self.name)

        _result = []

        # check for error in result
        # TODO: have the result output the error
        if result.error:
            data = {}
            data["Error"] = [f"Command: {result.command_code} incurred an error or errors during execution or processing", ""]
            data["Error Count"] = [len(result.error_messages), ""]
            for i, message in enumerate(result.error_messages):
                data[f"Error #{i}"] = [message, ""]

        if len(result.readings) == 0:
            return _result

        displayData : list[Reading] = self.format_and_filter_data(result)
        log.debug(f"displayData: {displayData}")

        _result.append("<table><tr><th>Parameter</th><th>Value</th><th>Unit</th></tr>")
        for response in displayData:
            key = response.get_data_name()
            value = response.get_data_value()
            unit = response.get_data_unit()
            _result.append(f"<tr><td>{key}</td><td>{value}</td><td>{unit}</td></tr>")
        _result.append("</table>")
        return _result



import logging
from powermon.formats.abstractformat import AbstractFormat
from powermon.commands.result import Result

log = logging.getLogger("raw")


class raw(AbstractFormat):
    def __init__(self, formatConfig):
        super().__init__(formatConfig)
        self.name = "raw"
        self.extra_info = formatConfig.get("extra_info", False)
        
    # def set_command_description(self, command_description):
    #     pass

    def format(self, result: Result, device_info):
        log.info("Using output formatter: %s", self.name)

        data = result.raw_response
        log.debug(f"data: {data}")
        return [data]



import json as js
import logging
import re
from datetime import datetime
from time import sleep

from ..helpers import get_kwargs, key_wanted
from .mqtt import mqtt

log = logging.getLogger("hassd_mqtt")


class hassd_mqtt(mqtt):
    def __str__(self):
        return """outputs the to the supplied mqtt broker in hass format: eg "homeassistant/sensor/mpp_{tag}_{key}/state" """

    def __init__(self, *args, **kwargs) -> None:
        log.debug(f"__init__: kwargs {kwargs}")

    def build_msgs(self, *args, **kwargs):
        log.debug(f"kwargs {kwargs}")
        data = get_kwargs(kwargs, "data")
        # Clean data
        command = data.pop("_command", None)
        data.pop("_command_description", None)
        data.pop("raw_response", None)

        # check if config supplied
        config = get_kwargs(kwargs, "config")
        if config is not None:
            log.debug(f"config: {config}")
            # try for fullconfig
            fullconfig = get_kwargs(kwargs, "fullconfig")
            # get results topic
            # results_topic = config.get("results_topic", None)
            # get formatting info
            remove_spaces = config.get("remove_spaces", True)
            keep_case = config.get("keep_case", False)
            filter = config.get("filter", None)
            excl_filter = config.get("excl_filter", None)
            tag = config.get("tag", None)
            device = fullconfig.get("device", {})
            device_name = device.get("name", "mppsolar")
            device_id = device.get("id", "mppsolar")
            device_model = device.get("model", "mppsolar")
            device_manufacturer = device.get("manufacturer", "mppsolar")
        else:
            # results_topic = None
            # get formatting info
            remove_spaces = True
            keep_case = get_kwargs(kwargs, "keep_case")
            filter = get_kwargs(kwargs, "filter")
            excl_filter = get_kwargs(kwargs, "excl_filter")
            tag = get_kwargs(kwargs, "tag")
            device_name = get_kwargs(kwargs, "name", "mppsolar")
            device_id = device_name
            device_model = device_name
            device_manufacturer = "MPP-Solar"

        if filter is not None:
            filter = re.compile(filter)
        if excl_filter is not None:
            excl_filter = re.compile(excl_filter)
        if tag is None:
            if command:
                tag = command
            else:
                tag = "mppsolar"

        # Build array of mqtt messages with hass update format
        config_msgs = []
        value_msgs = []

        # Loop through responses
        for key, values in data.items():
            orig_key = key
            value = values[0]
            unit = values[1]
            icon = None
            if len(values) > 2 and values[2] and "icon" in values[2]:
                icon = values[2]["icon"]
            device_class = None
            if len(values) > 2 and values[2] and "device-class" in values[2]:
                device_class = values[2]["device-class"]
            state_class = None
            if len(values) > 2 and values[2] and "state_class" in values[2]:
                state_class = values[2]["state_class"]

            # remove spaces
            if remove_spaces:
                key = key.replace(" ", "_")
            if not keep_case:
                # make lowercase
                key = key.lower()
            if key_wanted(key, filter, excl_filter):
                #
                # CONFIG / AUTODISCOVER
                #
                # <discovery_prefix>/<component>/[<node_id>/]<object_id>/config
                # topic "homeassistant/binary_sensor/garden/config"
                # msg '{"name": "garden", "device_class": "motion", "state_topic": "homeassistant/binary_sensor/garden/state", "unit_of_measurement": "°C", "icon": "power-plug"}'

                # For binary sensors
                if unit == "bool" or value == "enabled" or value == "disabled":
                    sensor = "binary_sensor"
                    if value == 0 or value == "0" or value == "disabled":
                        # for QPIWS one can add [or tag == "myQPIWStag"], if there's a QPIWS section in mpp-solar.conf
                        value = "OFF"
                    elif value == 1 or value == "1" or value == "enabled":
                        value = "ON"
                else:
                    sensor = "sensor"
                topic = f"homeassistant/{sensor}/mpp_{tag}_{key}/config"
                topic = topic.replace(" ", "_")
                name = f"{tag} {orig_key}"
                payload = {
                    "name": f"{name}",
                    "state_topic": f"homeassistant/{sensor}/mpp_{tag}_{key}/state",
                    "unique_id": f"mpp_{tag}_{key}",
                    "force_update": "true",
                }
                if unit and unit != "bool":
                    payload["unit_of_measurement"] = f"{unit}"

                # payload["device"] = {"name": f"{device_name}", "identifiers": ["mppsolar"], "model": "PIP6048MAX", "manufacturer": "MPP-Solar"}
                payload["device"] = {
                    "name": device_name,
                    "identifiers": [device_id],
                    "model": device_model,
                    "manufacturer": device_manufacturer,
                }
                if device_class:
                    payload["device_class"] = device_class
                if state_class:
                    payload["state_class"] = state_class
                if icon:
                    payload.update({"icon": icon})
                if unit == "W":
                    payload.update({"state_class": "measurement", "device_class": "power"})
                if unit == "Wh" or unit == "kWh":
                    payload.update(
                        {
                            "icon": "mdi:counter",
                            "device_class": "energy",
                            "state_class": "total",
                            "last_reset": str(datetime.now()),
                        }
                    )

                # msg = {"topic": topic, "payload": payload, "retain": True}
                payloads = js.dumps(payload)
                # print(payloads)
                msg = {"topic": topic, "payload": payloads}
                config_msgs.append(msg)
                #
                # VALUE SETTING
                #
                # 'tag'/status/total_output_active_power/value 1250
                # 'tag'/status/total_output_active_power/unit W
                topic = f"homeassistant/{sensor}/mpp_{tag}_{key}/state"
                msg = {"topic": topic, "payload": value}
                value_msgs.append(msg)
        return config_msgs, value_msgs

    def output(self, *args, **kwargs):
        """Over write mqtt output as we want to send config msgs first...."""
        log.info("Using output processor: hassd_mqtt")
        log.debug(f"kwargs {kwargs}")
        data = get_kwargs(kwargs, "data")
        # exit if no data
        if data is None:
            return

        # get the broker instance
        mqtt_broker = get_kwargs(kwargs, "mqtt_broker")
        # exit if no broker
        if mqtt_broker is None:
            return

        # build the messages...
        config_msgs, value_msgs = self.build_msgs(**kwargs)
        log.debug(f"hassd_mqtt.output config_msgs {config_msgs}")
        log.debug(f"hassd_mqtt.output value_msgs {value_msgs}")

        # publish
        mqtt_broker.publishMultiple(config_msgs)
        sleep(0.5)
        mqtt_broker.publishMultiple(value_msgs)



""" outputs / api_mqtt.py """
import logging

# from powermon.dto.resultDTO import ResultDTO
from powermon.commands.result import Result
# from powermon.device import Device
from powermon.dto.commandDTO import CommandDTO
from powermon.dto.outputDTO import OutputDTO
from powermon.formats.simple import SimpleFormat
from powermon.outputs.abstractoutput import AbstractOutput

log = logging.getLogger("ApiMqtt")


class ApiMqtt(AbstractOutput):
    """ docstring about ApiMqtt """  # TODO: update docstring and __str__
    def __str__(self):
        return "outputs .... TODO"

    def __init__(self):
        super().__init__(name="ApiMqtt")
        self.topic_base : str = "powermon/"
        self.topic_type : str = "results/"

    def get_topic(self) -> str:
        return CommandDTO.get_command_result_topic().format(device_id=self.device_id, command_name=self.command_code)

    def process(self, command=None, result: Result=None, mqtt_broker=None, device_info=None):
        # exit if no data
        if result.raw_response is None:
            return

        # exit if no broker
        if mqtt_broker is None:
            log.error("No mqtt broker supplied")
            raise RuntimeError("No mqtt broker supplied")

        result_dto = result.to_dto()
        mqtt_broker.publish(self.get_topic(), result_dto.json())

    @classmethod
    def from_dto(cls, dto: OutputDTO) -> "ApiMqtt":
        formatter = SimpleFormat.from_dto(dto.format)
        api_mqtt = cls()
        api_mqtt.set_formatter(formatter)

    @classmethod
    def from_config(cls, output_config) -> "ApiMqtt":
        log.debug("config: %s", output_config)  # TODO: sort from_config
        return cls()



""" dto / commandDTO.py """
from pydantic import BaseModel
from .triggerDTO import TriggerDTO
from .outputDTO import OutputDTO


class CommandDTO(BaseModel):
    """ model/allowed elements for a command data transfer object """
    command_code: str
    device_id: str
    result_topic: None | str = None
    trigger: TriggerDTO
    outputs: list[OutputDTO]

    @classmethod
    def run_api_command(cls, command_code: str, device_id: str) -> "CommandDTO":
        # FIXME: whats going on here - needs moving
        command_dto : CommandDTO = CommandDTO(
            command_code=command_code,
            device_id=device_id,
            result_topic=cls.get_command_result_topic().format(device_id=device_id, command_name=command_code),
            trigger=TriggerDTO(
                trigger_type="once",
                value="0"
                ),
            outputs=[])
        return command_dto

    @classmethod
    def get_command_result_topic(cls) -> str:
        # FIXME: needs moving and fixing (to return config value)
        return "powermon/{device_id}/results/{command_name}"



""" formats / simple.py """
import logging
from powermon.formats.abstractformat import AbstractFormat
from powermon.dto.formatDTO import FormatDTO
from powermon.commands.result import Result
from powermon.commands.reading import Reading

log = logging.getLogger("simple")


class SimpleFormat(AbstractFormat):
    """ simple format - {name}={value}{unit} format """
    def __init__(self, formatConfig):
        super().__init__(formatConfig)
        self.name = "simple"
        self.extra_info = formatConfig.get("extra_info", False)

    # def set_command_description(self, command_description):
    #     pass

    def format(self, command, result: Result, device_info) -> list:

        _result = []

        # check for error in result
        if result.error:
            _result.append(f"Error Count: {len(result.error_messages)}")
            for i, message in enumerate(result.error_messages):
                _result.append(f"Error #{i}: {message}")
            # return _result

        if len(result.readings) == 0:
            return _result

        display_data : list[Reading] = self.format_and_filter_data(result)

        # build data to display
        for reading in display_data:
            name = reading.get_data_name()
            value = reading.get_data_value()
            unit = reading.get_data_unit()
            if self.extra_info:
                extra = ""
                if reading.get_device_class() is not None:
                    extra = " " + reading.get_device_class()
                if reading.get_icon() is not None:
                    extra += " " + reading.get_icon()
                if reading.get_state_class() is not None:
                    extra += " " + reading.get_state_class()
                _result.append(f"{name}={value}{unit}{extra}")
            else:
                _result.append(f"{name}={value}{unit}")
        return _result

    @classmethod
    def from_dto(cls, dto: FormatDTO):
        """ build class object from dto """
        return cls(formatConfig=dto)



""" reading.py """
from powermon.dto.reading_dto import ReadingDTO


class Reading:
    """ class to contain the raw reading, processed reading and reading definition for a particular reading """
    def __str__(self):
        return f"Reading: {self.data_name=}, {self.data_value=}, {self.data_unit=}, {self.is_valid=}"

    def __init__(self, data_name: str, data_value: str, data_unit: str, device_class: str = None, state_class: str = None, icon: str = None) -> None:
        self.data_name = data_name
        self.data_value = data_value
        self.data_unit = data_unit
        self.device_class = device_class
        self.icon = icon
        self.state_class = state_class
        self.is_valid = True

    def to_dto(self) -> ReadingDTO:
        """ convert Reading to a data transfer object """
        return ReadingDTO(data_name=self.get_data_name(), data_value=self.get_data_value(), data_unit=self.get_data_unit())

    @property
    def raw_value(self):
        """ the raw reading returned from the device - as returned """
        return self._raw_value

    @raw_value.setter
    def raw_value(self, value):
        self._raw_value = value

    @property
    def processed_value(self):
        """ the reading after processing and conversion - the reading we want to see """
        return self._processed_value

    @processed_value.setter
    def processed_value(self, value):
        # should take raw value and 'process it'
        self._processed_value = value

    @property
    def reading_definition(self):
        """ the reading definition associated with this reading """
        return self._reading_definition

    @reading_definition.setter
    def reading_definition(self, value):
        self._reading_definition = value

    # the below are / should be part of reading definition
    def get_data_name(self) -> str:
        return self.data_name.replace(" ", "_").lower()

    def get_data_unit(self) -> str:
        if self.data_unit is None:
            return ""
        return self.data_unit

    def get_data_value(self) -> str:
        return self.data_value

    def get_icon(self) -> str | None:
        return self.icon

    def get_device_class(self) -> str | None:
        return self.device_class

    def get_state_class(self) -> str | None:
        return self.state_class



""" outputs / __init__.py """
import logging
from enum import auto

from strenum import LowercaseStrEnum

from powermon.formats import DEFAULT_FORMAT, getFormatfromConfig

# Set-up logger
log = logging.getLogger("outputs")

DEFAULT_OUTPUT = "screen"


class OutputType(LowercaseStrEnum):
    """ enum of valid output types """
    SCREEN = auto()
    MQTT = auto()
    API_MQTT = auto()


def getOutputClass(output_type, formatter, output_config={}):
    output_class = None
    # Only import the required class
    log.debug("outputType %s", output_type)
    if output_type == OutputType.MQTT:
        from powermon.outputs.mqtt import MQTT
        output_class = MQTT.from_config(output_config)
        output_class.formatter = formatter
    elif output_type == OutputType.API_MQTT:
        from powermon.outputs.api_mqtt import ApiMqtt
        output_class = ApiMqtt.from_config(output_config)
        output_class.formatter = formatter
    else:
        from powermon.outputs.screen import Screen
        output_class = Screen.from_config(output_config)
        output_class.formatter = formatter
    return output_class


def multiple_from_config(outputs_config):
    """ return one or more output classes from a config """
    # outputs can be None,
    # str (eg screen),
    # list (eg [{'type': 'screen', 'format': 'simple'}, {'type': 'screen', 'format': {'type': 'htmltable'}}])
    # dict (eg {'format': 'table'})
    # print("outputs %s, type: %s" % (outputs, type(outputs)))
    _outputs = []
    log.debug("processing outputs_config: %s", outputs_config)
    if outputs_config is None:
        _outputs.append(parse_output_config({"type": "screen", "format": DEFAULT_FORMAT}))
    elif isinstance(outputs_config, str):
        # eg 'screen'
        _outputs.append(parse_output_config({"type": outputs_config, "format": DEFAULT_FORMAT}))
    elif isinstance(outputs_config, list):
        # eg [{'type': 'screen', 'format': 'simple'}, {'type': 'screen', 'format': {'type': 'htmltable'}}]
        for output_config in outputs_config:
            _outputs.append(parse_output_config(output_config))
    elif isinstance(outputs_config, dict):
        # eg {'format': 'table'}
        _outputs.append(parse_output_config(outputs_config))
    else:
        pass
    return _outputs


def parse_output_config(output_config):
    """ generate a single output object from a config """
    log.debug("parse_output_config, config: %s", output_config)
    log.debug("got dict type output_config: %s", output_config)
    output_type = output_config.get("type", DEFAULT_OUTPUT)
    format_config = output_config.get("format", DEFAULT_FORMAT)
    _format = getFormatfromConfig(format_config)
    log.debug("got format: %s", (_format))
    _output = getOutputClass(output_type, formatter=_format)
    log.debug("got output: %s", _output)
    return _output



""" serialport.py """
import logging
import time

import serial

from powermon.commands.command import Command
from powermon.commands.result import Result
from powermon.dto.portDTO import PortDTO
from powermon.ports.abstractport import AbstractPort
from powermon.protocols import get_protocol_definition
from powermon.protocols.ved import VictronCommandType

log = logging.getLogger("SerialPort")


class SerialPort(AbstractPort):
    """ serial port object - normally a usb to serial adapter """

    def __str__(self):
        return f"SerialPort: {self.path=}, {self.baud=}, protocol:{self.protocol}, {self.serial_port=}, {self.error_message=}"

    @classmethod
    def from_config(cls, config=None):
        log.debug("building serial port. config:%s", config)
        path = config.get("path", "/dev/ttyUSB0")
        baud = config.get("baud", 2400)
        # get protocol handler, default to PI30 if not supplied
        protocol = get_protocol_definition(protocol=config.get("protocol", "PI30"))
        return cls(path=path, baud=baud, protocol=protocol)

    def __init__(self, path, baud, protocol) -> None:
        super().__init__(protocol=protocol)
        self.path = path
        self.baud = baud
        self.serial_port = None
        # self.error_message = None

    def to_dto(self) -> PortDTO:
        dto = PortDTO(type="serial", path=self.path, baud=self.baud, protocol=self.protocol.to_dto())
        return dto

    def is_connected(self):
        return self.serial_port is not None and self.serial_port.is_open

    def connect(self) -> int:
        log.debug("usbserial port connecting. path:%s, baud:%s", self.path, self.baud)
        try:
            self.serial_port = serial.Serial(port=self.path, baudrate=self.baud, timeout=1, write_timeout=1)
            log.debug(self.serial_port)
        except ValueError as e:
            log.error("Incorrect configuration for serial port: %s", e)
            self.error_message = str(e)
            self.serial_port = None
        except serial.SerialException as e:
            log.error("Error opening serial port: %s", e)
            self.error_message = str(e)
            self.serial_port = None
        return self.is_connected()

    def disconnect(self) -> None:
        log.debug("usbserial port disconnecting")
        if self.serial_port is not None:
            self.serial_port.close()
        self.serial_port = None

    def send_and_receive(self, command: Command) -> Result:
        full_command = command.full_command
        response_line = None
        log.debug("port: %s, full_command: %s", self.serial_port, full_command)
        if not self.is_connected():
            raise RuntimeError("Serial port not open")
        try:
            log.debug("Executing command via usbserial...")
            self.serial_port.reset_input_buffer()
            self.serial_port.reset_output_buffer()
            if full_command == VictronCommandType.LISTEN:
                # this command type doesnt need to send a command, it just listens on the serial port
                response_line = b""
                for _ in range(30):
                    _response = self.serial_port.read_until(b"\n")
                    response_line += _response
            else:
                c = self.serial_port.write(full_command)
                log.debug("Wrote %i bytes", c)
                self.serial_port.flush()
                time.sleep(0.1)  # give serial port time to receive the data
                response_line = self.serial_port.read_until(b"\r")
            log.debug("serial response was: %s", response_line)
            # response = self.get_protocol().check_response_and_trim(response_line)
            result = command.build_result(raw_response=response_line, protocol=self.protocol)
            return result
        except Exception as e:
            log.warning("Serial read error: %s", e)
            result.error = True
            result.error_messages.append(f"Serial read error {e}")
            self.disconnect()
            return result



import binascii
import json as js
import logging
import time

import paho.mqtt.client as mqttc

from ..helpers import get_kwargs
from .baseio import BaseIO

log = logging.getLogger("MqttIO")


class MqttIO(BaseIO):
    def __init__(self, *args, **kwargs) -> None:
        # self._serial_port = device_path
        # self._serial_baud = serial_baud
        self.mqtt_broker = get_kwargs(kwargs, "mqtt_broker", "localhost")
        self.mqtt_port = self.mqtt_broker.port
        self.mqtt_user = self.mqtt_broker.username
        self.mqtt_pass = self.mqtt_broker.password
        # self.mqtt_port = get_kwargs(kwargs, "mqtt_port", 1883)
        # self.mqtt_user = get_kwargs(kwargs, "mqtt_user")
        # self.mqtt_pass = get_kwargs(kwargs, "mqtt_pass")
        self.client_id = get_kwargs(kwargs, "client_id")
        log.info(
            f"__init__: client_id: {self.client_id}, mqtt_broker: {self.mqtt_broker}, port: {self.mqtt_port}, user: {self.mqtt_user}, pass: {self.mqtt_pass}"
        )
        self._msg = None

    def sub_cb(self, client, userdata, message):
        log.debug(f"Mqttio sub_cb got msg, topic: {message.topic}, payload: {message.payload}")
        self._msg = message

    def send_and_receive(self, *args, **kwargs) -> dict:
        full_command = get_kwargs(kwargs, "full_command")
        command = get_kwargs(kwargs, "command")
        client_id = self.client_id

        wait_time = 5
        # response_line = None
        command_topic = f"{client_id}/command"
        result_topic = f"{client_id}/result"
        # print(self.mqtt_broker)
        # Create mqtt client
        # Client(client_id="", clean_session=True, userdata=None, protocol=MQTTv311, transport="tcp")
        mqtt_client = mqttc.Client()
        # mqtt_client.on_connect = on_connect

        if self.mqtt_user is not None and self.mqtt_pass is not None:
            # auth = {"username": self.mqtt_user, "password": self.mqtt_pass}
            log.info(
                f"Using mqtt authentication, username: {self.mqtt_user}, password: [supplied]"
            )
            mqtt_client.username_pw_set(self.mqtt_user, password=self.mqtt_pass)
        else:
            log.debug("No mqtt authentication used")
            # auth = None

        mqtt_client.connect(self.mqtt_broker, port=self.mqtt_port)

        command_hex = binascii.hexlify(full_command)
        payload = {"command": command, "command_hex": command_hex.decode()}
        payload = js.dumps(payload)

        log.debug(f"Publishing {payload} to topic: {command_topic}")

        mqtt_client.publish(command_topic, payload=payload)

        mqtt_client.on_message = self.sub_cb
        mqtt_client.subscribe(result_topic)
        mqtt_client.loop_start()
        time.sleep(wait_time)
        mqtt_client.loop_stop(force=False)

        if self._msg is None:
            # Didnt get a result
            return {
                "ERROR": [
                    f"Mqtt result message not received on topic {result_topic} after {wait_time}sec",
                    "",
                ]
            }
        else:
            msg_topic = self._msg.topic
            # decode the payload
            # payload should be a json dumped byte string
            # payload: b'{"command_hex": "515049beac0d", "result": "", "command": "QPI"}'
            log.debug(
                f"mqtt raw response on {self._msg.topic} was: {self._msg.payload}, payload type: {type(self._msg.payload)}"
            )
            # Return the byte-string to a dict
            payload_dict = js.loads(self._msg.payload)
            # Get 'results', and convert back to bytes
            result = binascii.unhexlify(payload_dict["result"])
            # TODO: Currently ignoring this - might want to update return types at some point
            cmd = payload_dict["command"]
            self._msg = None
            log.debug(f"mqtt response on {msg_topic} for command {cmd} was: {result}")
            return result
